/***************************************************************************
*   Copyright (C) 2003 by Graham Dennis                                   *
*   u3952328@anu.edu.au                                                   *
*                                                                         *
*   Permission is hereby granted, free of charge, to any person obtaining *
*   a copy of this software and associated documentation files (the       *
*   "Software"), to deal in the Software without restriction, including   *
*   without limitation the rights to use, copy, modify, merge, publish,   *
*   distribute, sublicense, and/or sell copies of the Software, and to    *
*   permit persons to whom the Software is furnished to do so, subject to *
*   the following conditions:                                             *
*                                                                         *
*   The above copyright notice and this permission notice shall be        *
*   included in all copies or substantial portions of the Software.       *
*                                                                         *
*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
*   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
*   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
*   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
*   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
*   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
*   OTHER DEALINGS IN THE SOFTWARE.                                       *
***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <cstdlib>


#include "common.h"

using namespace Rendezvous;

void dlocate( const char*const file, const char* const function, int line )
{
	printf( "File: \"%s\" Function: %s() Line: %d\n", file, function, line );
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//								class MemorySeg											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

MemorySeg::MemorySeg( MemorySeg &MemSeg )
{
	if(MemSeg.Base)
		Base = memcpy(malloc(MemSeg.BufUsed), MemSeg.Base, MemSeg.BufUsed);
	else
		Base = NULL;
	BufSize = MemSeg.BufSize;
	BufUsed = MemSeg.BufUsed;
	MinSizeInc = MemSeg.MinSizeInc;
}

MemorySeg::MemorySeg( size_t nBufSize, size_t nMinSizeInc )
{
	MinSizeInc = nMinSizeInc;
	Base = NULL;
	BufUsed=0;
	BufSize=0;
	if ( nBufSize != 0 )
		Alloc( nBufSize );
	if ( !Base ) { BufSize = BufUsed = 0; }
};

MemorySeg::~MemorySeg()
{
	Free();
}

void* MemorySeg::Alloc( size_t nbytes )
{
	void * ret;
	debug_locate();
	if ( MinSizeInc == 0 )
		MinSizeInc = MemorySeg::DefaultMinSizeInc;
	if ( Base == NULL ) {
		debug_locate();
		BufSize = MAX( nbytes, MinSizeInc );
		BufUsed = 0;
		Base = memset( malloc( BufSize ), 0, BufSize );
	} else if ( BufUsed + nbytes > BufSize ) {
		BufSize += MAX( nbytes, MinSizeInc );
		Base = realloc( Base, BufSize );
	}
	 ret = (void*)(ptrdiff_t( Base ) + BufUsed);
	BufUsed += nbytes;
	return ret;
}

void MemorySeg::Free()
{
	if(Base)
		free( Base );
	BufSize = 0;
	BufUsed = 0;
	Base = NULL;
}

void MemorySeg::FreeFirst( size_t nbytes )
{
	memmove( Base, (char*)Base + nbytes, BufUsed - nbytes );
	BufUsed -= nbytes;
	BufSize -= nbytes;
	Base = realloc( Base, BufSize );
}

int MemorySeg::Extract( size_t offset, void *buf, size_t nbytes )
{
	if ( buf == NULL ) return -1;
//	if ( nbytes < 0 ) return -1;
//	if ( offset < 0 ) return -1;
	if ( offset + nbytes > BufUsed ) return -1;
	memcpy( buf, (char*) Base + offset, nbytes );
	return 0;
}

int MemorySeg::Write( size_t offset, void* buf, size_t nbytes )
{
	if ( buf == NULL ) return -1;
//	if ( nbytes < 0 ) return -1;
//	if ( offset < 0 ) return -1;
	if ( offset + nbytes > BufUsed ) return -1;
	memcpy( (char*)Base + offset, buf, nbytes );
	return 0;
}

void *MemorySeg::Offset( size_t offset )
{
	return (char*)(Base) + offset;
}

int MemorySeg::ZeroMemory()
{
	memset(Base, 0, BufSize);
	return 0;
}

size_t MemorySeg::Size() { return BufUsed; }
size_t MemorySeg::Allocated() { return BufSize; }
void* MemorySeg::BasePtr() { return Base; }
MemorySeg::operator char*() { return (char*)( Base); }
MemorySeg::operator void*() { return Base; }

const MemorySeg& MemorySeg::operator=(const MemorySeg& mem)
{
	Free();
	if(mem.Base)
		Base = memcpy(malloc(mem.BufUsed), mem.Base, mem.BufUsed);
	else
		Base = NULL;
	BufSize = mem.BufSize;
	BufUsed = mem.BufUsed;
	MinSizeInc = mem.MinSizeInc;
	return mem;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//								class FDSet												//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

FDSet::FDSet() { Zero(); fdmax = 0;}

FDSet::FDSet( FDSet & old ) { fdSet = old.fdSet ; fdmax = old.fdmax;}

FDSet FDSet::operator= ( FDSet& old )
{
	fdSet = old.fdSet;
	fdmax = old.fdmax;
	return old;
}

int FDSet::IsSet( const int fd )
{
	assert(fd>=0);
	return FD_ISSET( fd, &fdSet );
}

void FDSet::Set( const int fd )
{
	assert( fd != -1 );
	FD_SET( fd, &fdSet );
	if ( fd >= fdmax )
		fdmax = fd + 1;
}

void FDSet::Unset( const int fd )
{
	assert( fd != -1 );
	FD_CLR( fd, &fdSet );
	if ( fd + 1 == fdmax ) {
		int newmax = 0;
		for ( int i = 0;i < fdmax;i++ ) {
			if ( FD_ISSET( i, &fdSet ) ) newmax = i;
		}
		fdmax = newmax + 1;
	}
}

void FDSet::Add( fd_set* fds, int fdmax )
{
	for ( int i = 0;i < fdmax;i++ ) {
		if ( FD_ISSET( i, fds ) )
			Set( i );
	}

}

void FDSet::Zero()
{
	FD_ZERO( &fdSet );
	fdmax = 0;
}

fd_set* FDSet::SetPtr() { return & fdSet; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//								class FDGroup											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

FDGroup::FDGroup()
{
	memset( &timeout, 0, sizeof( timeout ) );
	timeout.tv_sec = FDGroup::DefaultTimeout;
}

FDGroup FDGroup::operator= ( FDGroup& old )
{
	Read = old.Read;
	Write = old.Write;
	Except = old.Except;
	return old;
}

int FDGroup::MaxFD()
{
	return MAX( MAX( Read.fdmax, Write.fdmax ), Except.fdmax );
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//								class Message											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Message::Message(Message * msg)
{
	HashMapIterator i = msg->Data.begin();
	int j=msg->Data.size();
	while(j--) {
		AddData((*i).first, (*i).second.second, (*i).second.first);
		++i;
	}

}

Message::Message(Message& msg)
{
	HashMapIterator i = msg.Data.begin();
	int j=msg.Data.size();
	while(j--) {
		AddData((*i).first, (*i).second.second, (*i).second.first);
		++i;
	}
}

Message::Message() {Data.clear();}

Message::~Message()
{
	HashMapIterator i = Data.begin(), temp;
	int j = Data.size();
	char *string;
	while ( j-- ) {
		temp = i++;
		string = (char*) (*temp).first;
		( *temp ).second.first = 0;
		if ( ( *temp ).second.second != NULL )
			free( ( *temp ).second.second );
		if(string)
			free(string);
	}
	Data.clear();
}

void Message::AddData( const char* Key, const void *Buffer, size_t BufSize, Message::Options alloc )
{
	char* string = strdup(Key);
	pair<size_t, void*> Element;
	Element.first = BufSize;
	if(alloc!=DoNotDuplicate)
		Element.second = memcpy( malloc( BufSize ), Buffer, BufSize );
	else
		Element.second = (void*)Buffer;
	Data[ string ] = Element;
}

void Message::Add( const char* Key, MemorySeg& MemSeg ) { AddData( Key, MemSeg.BasePtr(), MemSeg.Size() ); }
void Message::Add( const char* Key, MemorySeg* MemSeg ) { Add( Key, *MemSeg ); }

void Message::Add( const char* Key, const char* Data, Message::Options term)
{
	AddData( Key, (void*) Data , strlen( Data ) + (term == Message::NoNullTerminator? 0 : 1) );
}

void Message::AddKey( const char* NewKey, const char * OldKey, Message* m )
{
	if ( m->Data.count( OldKey ) == 0 ) return ;
	AddData( NewKey, m->Data[ OldKey ].second, m->Data[ OldKey ].first );
}

void Message::AddKey( const char* Key, Message *m ) { AddKey( Key, Key, m ); }

void Message::Add( const char*Key, const unsigned char&value ) { Addu8( Key, value ); }
void Message::Add( const char*Key, const signed char& value ) { Addi8( Key, value ); }
void Message::Add( const char*Key, const unsigned int &value ) {Addu16( Key, value ); }
void Message::Add( const char*Key, const signed int &value ) {Addi16( Key, value ); }
void Message::Add( const char*Key, const unsigned long &value ) {Addu32( Key, value ); }
void Message::Add( const char *Key,const  signed long &value ) {Addi32( Key, value ); }

void Message::Addu8(const char*Key, const uint8_t&value) { Add_internal(Key, "%u", value); }
void Message::Addi8(const char*Key, const int8_t&value) { Add_internal(Key, "%i", value); }
void Message::Addu16(const char*Key, const uint16_t& value) {Add_internal(Key, "%u", value); }
void Message::Addi16(const char*Key, const int16_t& value) {Add_internal(Key, "%i", value); }
void Message::Addu32(const char*Key, const uint32_t& value) {Add_internal(Key, "%lu", value); }
void Message::Addi32(const char*Key, const int32_t& value) {Add_internal(Key, "%li", value); }

void Message::Function( const char* func )
{
	Add( "Function", func );
}

const char* Message::Function()
{
	if ( Data.count( "Function" ) == 0 ) return NULL;
	return (const char*)Data[ "Function" ].second;
}

void Message::Callback( const char* type)
{
	Function("Callback");
	Add("CallbackType", type);
}

const char* Message::CallbackType()
{
	if(strcmp(Function(), "Callback")!=0) return NULL;
	if(Data.count("CallbackType") ==0) return NULL;
	return (const char*) Data[ "CallbackType"].second;
}
