 /*
  * ichatcontact.cpp  -  Base class for the Kopete iChat protocol contact
  *
  * Copyright (c) 2003 by Graham Dennis <u3952328@anu.edu.au>
  * Copyright (c) 2002-2003 by Till Gerken <till@tantalo.net>
  * Copyright (c) 2002 by Daniel Stone <dstone@kde.org>
  *
  * Kopete    (c) by the Kopete developers  <kopete-devel@kde.org>
  *
  * *************************************************************************
  * *                                                                       *
  * * This program is free software; you can redistribute it and/or modify  *
  * * it under the terms of the GNU General Public License as published by  *
  * * the Free Software Foundation; either version 2 of the License, or     *
  * * (at your option) any later version.                                   *
  * *                                                                       *
  * *************************************************************************
  */

#include <qapplication.h>
#include <qcursor.h>
#include <qfont.h>
#include <qptrlist.h>
#include <qstrlist.h>
#include <qdatetime.h>
#include <qimage.h>

#include <kaction.h>
#include <kconfig.h>
#include <kdebug.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <ksockaddr.h>

#include "stream.h"
#include "message.h"
//#include "types.h"

#include "dlgichatvcard.h"
#include "ichatcontact.h"
#include "ichatprotocol.h"
//#include "ichatresource.h"
#include "kopetestdaction.h"
#include "kopetemessage.h"
#include "kopetemessagemanager.h"
#include "kopetemessagemanagerfactory.h"
#include "kopeteview.h"
#include "kopetemetacontact.h"
#include "kopetegroup.h"
#include "kopetecontactlist.h"

/**
 * iChatContact constructor
 */
iChatContact::iChatContact (QString userId, RendezvousClient::ServiceInfo si, iChatAccount * p, KopeteMetaContact * mc)
				: KopeteContact (p, userId.lower(), mc)
{
	parentMetaContact = mc;
	messageManager = 0L;
	kdDebug(ICHAT_DEBUG_GLOBAL) << "UserId = " << userId << endl;
	UserID = userId;
	Domain = si.Domain;
	isLocalTyping=false;
	v4Addr = v6Addr = NULL;
	IdleAt=0;
	if(si.Type == QString::fromLatin1("_ichat._tcp.") ) {
		kdDebug(ICHAT_DEBUG_GLOBAL) << "Service type is _ichat: " << si.Type << endl;
		Service = iChatContact::_ichat;
	} else if (si.Type == QString::fromLatin1("_presence._tcp.")) {
		kdDebug(ICHAT_DEBUG_GLOBAL) << "Service type is _presence: " << si.Type << endl;
		Service = iChatContact::_presence;
	} else {
		Service = iChatContact::UnknownService;
		kdDebug (ICHAT_DEBUG_GLOBAL) << "We have one heck of a problem here. The service " << si.Type << "is unknown" << endl;
	}
	FQDN = si.FQDN;
	mResolve = NULL;
	mNullQuery = NULL;
	memset(phsh, 0, sizeof(phsh));
	slotUpdatePresence (static_cast<iChatProtocol *>(protocol())->iChatKOSOffline, QString::null);
	slotUpdateContact(si);

	// specifically cause this instance to update this contact as offline

//	connect(this, SIGNAL(displayNameChanged(const QString &, const QString &)), this, SLOT(slotRenameContact(const QString &, const QString &)));

}

/**
 * Another iChatContact constructor -- this should be removed
 */
iChatContact::iChatContact (QString userId, QString nickname, iChatAccount * p, KopeteMetaContact * mc)
				: KopeteContact (p, userId.lower(), mc)
{
	parentMetaContact = mc;
	isLocalTyping=false;

	messageManager = 0L;
	UserID = userId;
	Domain = QString::fromLatin1("local.");
	Service = iChatContact::UnknownService;
	FQDN = QString::null;
	setDisplayName(nickname);
	mResolve = NULL;
	mNullQuery = NULL;
	v4Addr = v6Addr = NULL;
	IdleAt=0;
	memset(phsh, 0, sizeof(phsh));

	// specifically cause this instance to update this contact as offline
	slotUpdatePresence (static_cast<iChatProtocol *>(protocol())->iChatKOSOffline, QString::null);

//	connect(this, SIGNAL(displayNameChanged(const QString &, const QString &)), this, SLOT(slotRenameContact(const QString &, const QString &)));

}


/* Return the user ID */
QString iChatContact::userId () const
{
	return UserID;
}

void iChatContact::slotIncomingConversation(int sock, const QString &host, int port)
{
	if(!ChatStream.isActive()) {
		QObject::connect( &ChatStream, SIGNAL(connected()), this, SLOT(slotStreamConnect()) );
		QObject::connect( &ChatStream, SIGNAL(handshaken()), this, SLOT(slotStreamHandshake()));
		QObject::connect( &ChatStream, SIGNAL(error(const Jabber::StreamError& )), this, SLOT(slotStreamError(const Jabber::StreamError& )));
		QObject::connect(&ChatStream, SIGNAL(closeFinished()), this, SLOT(slotStreamClosed()));
		QObject::connect(&ChatStream, SIGNAL(receivePacket(const QDomElement& )), this, SLOT(slotStreamRecieve(const QDomElement&)));
		ChatStream.setSocket(sock, host, port);
		manager(true);
	} else {
		kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] We had an incoming connection, but we are apparently already connected to them" << endl;
	}
}


KopeteMessageManager *iChatContact::manager (bool canCreate)
{

	// create a new message manager if there is none
	if (!messageManager && canCreate)
	{
		kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Creating new message manager." << endl;

		KopeteContactPtrList contactList;

		contactList.append (this);

		messageManager = KopeteMessageManagerFactory::factory ()->create (account()->myself (), contactList, protocol());

		QObject::connect (messageManager, SIGNAL (destroyed ()), this, SLOT (slotMessageManagerDeleted ()));
		QObject::connect (messageManager, SIGNAL (messageSent (KopeteMessage &, KopeteMessageManager *)), this, SLOT (slotSendMessage (KopeteMessage &)));
		QObject::connect(messageManager, SIGNAL(typingMsg(bool )), this, SLOT(slotTypingMsg(bool)));
		if(!ChatStream.isActive()) {	// TODO: Should we connect via IPv4 or IPv6 first?
			kdDebug (ICHAT_DEBUG_GLOBAL) << "Looks like we aren't connected... " << endl;

			if(v4Addr || v6Addr) {
				QObject::connect( &ChatStream, SIGNAL(connected()), this, SLOT(slotStreamConnect()) );
				QObject::connect( &ChatStream, SIGNAL(handshaken()), this, SLOT(slotStreamHandshake()));
				QObject::connect( &ChatStream, SIGNAL(error(const Jabber::StreamError& )), this, SLOT(slotStreamError(const Jabber::StreamError& )));
				QObject::connect(&ChatStream, SIGNAL(closeFinished()), this, SLOT(slotStreamClosed()));
				QObject::connect(&ChatStream, SIGNAL(receivePacket(const QDomElement& )), this, SLOT(slotStreamRecieve(const QDomElement&)));
				if(v4Addr) {
					kdDebug (ICHAT_DEBUG_GLOBAL) << "Connecting to...... " << v4Addr->pretty() << endl;
					ChatStream.connectToHost(v4Addr->nodeName(), Port);
				}else if(v6Addr) {
					kdDebug (ICHAT_DEBUG_GLOBAL) << "Connecting to remote host via..... IPv6?????? dest: "<< v6Addr->pretty() << endl;
					ChatStream.connectToHost(v6Addr->nodeName(), Port);
				}
			} else {
				kdDebug (ICHAT_DEBUG_GLOBAL) << "Both IPv4 and IPv6 are NULL!"<< endl;
			}
		} else {
			kdDebug (ICHAT_DEBUG_GLOBAL) << "Apparently we are active???? ==> Perhaps due to the other person establishing the connection" << endl;
		}
	}

	return static_cast < KopeteMessageManager * >(messageManager);

}

void iChatContact::slotStreamConnect()
{
	kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Stream is CONNECTED!!!!" << endl;
}

void iChatContact::slotStreamHandshake()
{
	kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Successful handshake!" << endl;
}

void iChatContact::slotStreamError( const Jabber::StreamError& err)
{
	kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Stream Error: " << err.toString() << endl;
	if (err.type() == Jabber::StreamError::Disconnected) {
		slotStreamClosed();
	}
}

void iChatContact::slotStreamClosed()
{
	kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Stream closed!" << endl;
		QObject::disconnect( &ChatStream, SIGNAL(connected()), this, SLOT(slotStreamConnect()) );
		QObject::disconnect( &ChatStream, SIGNAL(handshaken()), this, SLOT(slotStreamHandshake()));
		QObject::disconnect( &ChatStream, SIGNAL(error(const Jabber::StreamError& )), this, SLOT(slotStreamError(const Jabber::StreamError& )));
		QObject::disconnect(&ChatStream, SIGNAL(closeFinished()), this, SLOT(slotStreamClosed()));
		QObject::disconnect(&ChatStream, SIGNAL(receivePacket(const QDomElement& )), this, SLOT(slotStreamRecieve(const QDomElement&)));
	if(ChatStream.isActive()) {
		ChatStream.close();
	}
	if(manager()) {
	kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Manager is still here...!" << endl;
		KopeteMessage ClosedWindowMessage ( this, account()->myself(), i18n("%1 has closed the chat").arg(Name),
				  KopeteMessage::Internal,
				  KopeteMessage::PlainText, KopeteMessage::Chat);
		manager()->appendMessage(ClosedWindowMessage);
	}

}

void iChatContact::slotStreamRecieve( const QDomElement& dom)
{
	QString data;
	QTextStream stream(&data, IO_WriteOnly) ;
	stream << dom;
	kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Message received:" <<endl;
	kdDebug (ICHAT_DEBUG_GLOBAL) << data << endl;
	Jabber::Message jm;
	if(!jm.fromXml(dom))
		return;
	slotReceivedMessage(jm);
}

void iChatContact::slotMessageManagerDeleted ()
{

	kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Message manager has been deleted." << endl;
	messageManager = 0L;
	slotStreamClosed();

}


/* Return the reason why we are away */
QString iChatContact::reason () const
{
	return awayReason;
}

/* Return if we are reachable (defaults to true because
   we can send on- and offline */
bool iChatContact::isReachable ()
{
	return isOnline();
}

QString iChatContact::getv4Addy() const { return v4Addr ? v4Addr->nodeName() : QString::null; }
QString iChatContact::getv6Addy() const { return v6Addr ? v6Addr->nodeName() : QString::null; }


KActionCollection *iChatContact::customContextMenuActions ()
{
/*	KActionCollection *actionCollection = new KActionCollection (this);

//	new KAction (i18n ("(Re)send Authorization To"), "mail_forward", 0, this, SLOT (slotSendAuth ()), actionCollection, "actionSendAuth");
//	new KAction (i18n ("(Re)request Authorization From"), "mail_reply", 0, this, SLOT (slotRequestAuth ()), actionCollection, "actionRequestAuth");
	// FIXME wrong position in the popup menu
//	new KAction (i18n ("Remove Authorization From"), "mail_delete", 0, this, SLOT (slotRemoveAuth ()), actionCollection, "actionRemoveAuth");

//	KActionMenu *actionSetAvailability = new KActionMenu (i18n ("Set Availability"), "kopeteavailable", actionCollection, "ichat_online");

	actionSetAvailability->insert(new KAction (i18n ("Online"),         static_cast<iChatProtocol *>(protocol())->iChatKOSOnline.iconFor(this), 0, this, SLOT (slotStatusOnline ()), actionSetAvailability, "actionOnline"));
	actionSetAvailability->insert(new KAction (i18n ("Away"),           static_cast<iChatProtocol *>(protocol())->iChatKOSAway.iconFor(this), 0, this, SLOT (slotStatusAway ()), actionSetAvailability, "actionAway"));
	actionSetAvailability->insert(new KAction (i18n ("Do Not Disturb"), static_cast<iChatProtocol *>(protocol())->iChatKOSDND.iconFor(this), 0, this, SLOT (slotStatusDND ()), actionSetAvailability, "actionDND"));
	actionSetAvailability->insert(new KAction (i18n ("Invisible"),      static_cast<iChatProtocol *>(protocol())->iChatKOSInvisible.iconFor(this), 0, this, SLOT (slotStatusInvisible ()), actionSetAvailability, "actionInvisible"));

	KGlobal::config ()->setGroup ("iChat");

	return actionCollection;
*/
	return NULL;
}

void iChatContact::slotUpdatePresence (const KopeteOnlineStatus & newStatus, const QString & reason)
{
	awayReason = reason;
	setOnlineStatus (newStatus);
	if(newStatus == static_cast<iChatProtocol*>(protocol())->iChatKOSOffline)
	{
		mResolve=NULL;
		delete mNullQuery;
		mNullQuery=NULL;
	}
}

void iChatContact::slotUpdateContact(const RendezvousClient::ServiceInfo si)
{
	QMap<QString, QString> TXTMap;
	if(FQDN.isEmpty())
		FQDN = si.FQDN;
	if(Service == UnknownService) {
		if(FQDN.contains("_presence._tcp"))
			Service = _presence;
		else if (FQDN.contains("_ichat._tcp"))
			Service = _ichat;
		else
			kdDebug (ICHAT_DEBUG_GLOBAL) << "the service is unknown!!!" << endl;
	}
	if(Service == _presence)
		RendezvousClient::Action::ConvertTXT2Map(si.TXTFields, TXTMap);
	else {
		DecodeiChatTXTFields(si.TXTFields, TXTMap);
	}

	slotUpdateContact(si, TXTMap);
}

void iChatContact::slotUpdateContact (const RendezvousClient::ServiceInfo si, const QMap<QString, QString>& map)
{
	if(FQDN.isEmpty())
		FQDN = si.FQDN;
	if(Service == UnknownService) {
		if(FQDN.contains("_presence._tcp"))
			Service = _presence;
		else if (FQDN.contains("_ichat._tcp"))
			Service = _ichat;
		else
			kdDebug (ICHAT_DEBUG_GLOBAL) << "the service is unknown!!!" << endl;
	}
	if(si.isAddrV4)
		v4Addr = new KInetSocketAddress(si.Address, si.Port);
	else
		v6Addr = new KInetSocketAddress(si.Address, si.Port);
	Port = si.Port;
	mResolve = si.Resolver;
	TXTFields.clear();
	TXTFields = map;
	if( TXTFields.contains( QString::fromLatin1( "slumming" ))) {
		if( TXTFields[QString::fromLatin1("slumming")] == QString::fromLatin1("1")) {
			// We're slumming so this contact is duplicated
			static_cast<iChatAccount*>(account()) ->removeResolver(mResolve);
			mResolve = NULL;
			delete mNullQuery;
			mNullQuery=NULL;
			deleteLater();
			return;
		}
	}

	if ( TXTFields.contains( QString::fromLatin1( "1st" ) ) ) {
		Name = TXTFields[ QString::fromLatin1( "1st" ) ];
		if ( TXTFields.contains( QString::fromLatin1( "last" ) ) ) {
			if ( TXTFields[ QString::fromLatin1( "last" ) ] != " " && !TXTFields[ QString::fromLatin1("last")].isEmpty())
				Name += " " + TXTFields[ QString::fromLatin1( "last" ) ];
		}
		setDisplayName(Name);
	} else {
		setDisplayName(si.Name);
	}
	if ( TXTFields.contains( QString::fromLatin1("port.p2pj") ) )
		Port = TXTFields[ QString::fromLatin1("port.p2pj")].toUInt();
	if ( TXTFields.contains( "status" ) ) {
		if ( TXTFields.contains( "msg" ) )
			awayReason = TXTFields[ "msg" ];
		else
			awayReason = QString::null;
		if ( TXTFields[ "status" ] == "dnd" ) {
			slotUpdatePresence( static_cast<iChatProtocol *>( protocol() ) ->iChatKOSDND, awayReason );
		} else {
			if ( TXTFields[ "status" ] == "away" ) {
				slotUpdatePresence( static_cast<iChatProtocol *>( protocol() ) ->iChatKOSIdle, awayReason );
			} else if ( TXTFields[ "status" ] == "avail" ) {
				slotUpdatePresence( static_cast<iChatProtocol *>( protocol() ) ->iChatKOSOnline, awayReason );
			}
		}
	}
#define GMT1_1_2001       978307200
#define UNIX2OSX(x)   ((x)-GMT1_1_2001)
#define OSX2UNIX(x)   ((x)+GMT1_1_2001)

	if (TXTFields.contains( QString::fromLatin1("away"))) {
		Q_UINT32 temp;
		bool ok;
		temp = TXTFields[ QString::fromLatin1("away")].toULong(&ok);
		if (ok==false || temp == 0) {
			IdleAt = QDateTime::currentDateTime().toTime_t();
			setIdleTime(1);
			emit idleStateChanged(this);
		} else if (IdleAt!=OSX2UNIX(temp)) {
			IdleAt = OSX2UNIX(temp);
			setIdleTime(QDateTime::currentDateTime().toTime_t() - IdleAt);
			emit idleStateChanged(this);
		}
		
	} else if (IdleAt!=0) {
		setIdleTime(0);
		emit idleStateChanged(this);
	}
#undef GMT1_1_2001
#undef UNIX2OSX
#undef OSX2UNIX

	if(TXTFields.contains(QString::fromLatin1("phsh"))) {
		if(!mNullQuery) {
			mNullQuery = new RendezvousClient::NullQuery(si.Resolver->getClient());
			QObject::connect(mNullQuery, SIGNAL(sigNullResult( RendezvousClient::NullQuery*, QString )), this,
					SLOT(slotNullResult( RendezvousClient::NullQuery*, QString)));
			mNullQuery->Start(FQDN);
		}
	} else {
		if (!BuddyImage.isNull())
			BuddyImage.resize(0,0);
		delete mNullQuery;
		mNullQuery = NULL;
	}
}

QString iChatContact::getIdleStr() const
{
	if(!TXTFields.contains("away"))
		return QString::null;
	Q_INT32 temp = QDateTime::currentDateTime().toTime_t()-IdleAt;
	Q_INT32 days = temp / (60*60*24);
	Q_INT32 hours = (temp % (60*60*24))/ (60*60);
	Q_INT32 minutes = (temp % (60*60)) / 60;
//	Q_INT32 seconds = temp % 60;
	if(days) // Idle in excess of 1 day
		return i18n("%1 d %2 h %3 m").arg(days).arg(hours).arg(minutes);
	else if(hours) // Idle in excess of 1 hour
		return i18n("%1 h %2 m").arg(hours).arg(minutes);
	else
		return i18n("%1 m").arg(minutes);
}

// length of string resulting from encoding
#define b64_encode_len(x) (4 * (((x) + 2) / 3))
static int b64_decode(unsigned char* str)
{
    unsigned char *cur, *start;
    int d, dlast, phase;
    unsigned char c;
    static int table[256] = {
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 00-0F */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 10-1F */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,62,-1,-1,-1,63,  /* 20-2F */
        52,53,54,55,56,57,58,59,60,61,-1,-1,-1,-1,-1,-1,  /* 30-3F */
        -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,  /* 40-4F */
        15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,  /* 50-5F */
        -1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,  /* 60-6F */
        41,42,43,44,45,46,47,48,49,50,51,-1,-1,-1,-1,-1,  /* 70-7F */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 80-8F */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 90-9F */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* A0-AF */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* B0-BF */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* C0-CF */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* D0-DF */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* E0-EF */
        -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1   /* F0-FF */
    };

    d = dlast = phase = 0;
    start = str;
    for (cur = str; *cur != '\0'; ++cur )
    {
        /* handle newlines as seperate chunks */
        if(*cur == '\n' || *cur == '\r')
        {
//            phase = dlast = 0;
            continue;
        }

        d = table[(int)*cur];
        if(d != -1)
        {
            switch(phase)
            {
            case 0:
                ++phase;
                break;
            case 1:
                c = ((dlast << 2) | ((d & 0x30) >> 4));
                *str++ = c;
                ++phase;
                break;
            case 2:
                c = (((dlast & 0xf) << 4) | ((d & 0x3c) >> 2));
                *str++ = c;
                ++phase;
                break;
            case 3:
                c = (((dlast & 0x03 ) << 6) | d);
                *str++ = c;
                phase = 0;
                break;
            }
            dlast = d;
        }
    }
    *str = '\0';
    return str - start;
}

void iChatContact::slotNullResult( RendezvousClient::NullQuery* n, QString Data)
{
	(void)n;
	if(Data.length()==0) return;
	Data = Data.section("<data>", 1).section("</data>", 0, 0).simplifyWhiteSpace().remove(' ');
	if(Data.length()==0) return;
	const char* latin1form = Data.latin1();
	if(latin1form == NULL) return;
	unsigned int datalen = strlen(latin1form);
	if((datalen % 4) !=0) {
		kdDebug (ICHAT_DEBUG_GLOBAL) << "datalen is: " << datalen << endl;
		kdDebug (ICHAT_DEBUG_GLOBAL) << "buffer is: " << endl;
		kdDebug (ICHAT_DEBUG_GLOBAL) << Data << endl;
		kdDebug (ICHAT_DEBUG_GLOBAL) << "end buffer" << endl;
		return;
	}
	unsigned char *buffer = new unsigned char[datalen+1];
	memset(buffer, 0, datalen+1);
	memcpy(buffer, Data.latin1(), datalen);
	datalen = b64_decode(buffer);
	bool result;
	result = BuddyImage.loadFromData(buffer, datalen);
	switch(result)
	{
		case false:
			kdDebug (ICHAT_DEBUG_GLOBAL) << "Ohhhh, no fair! we couldn't load the image" << endl;
			break;
	}
	delete buffer;
}

void iChatContact::DecodeiChatTXTFields(const QStringList& List, QMap<QString, QString>& Map)
{
	QString Data = List.join(QString::null);
	Data = Data.section("<data>", 1).section("</data>", 0, 0).simplifyWhiteSpace().remove(' ').remove('\x00');
	long datalen = Data.length();
	if((datalen % 4) !=0) {
		kdDebug (ICHAT_DEBUG_GLOBAL) << "datalen is: " << datalen << endl;
		kdDebug (ICHAT_DEBUG_GLOBAL) << "buffer is: " << endl;
		kdDebug (ICHAT_DEBUG_GLOBAL) << Data << endl;
		kdDebug (ICHAT_DEBUG_GLOBAL) << "end buffer" << endl;
		return;
	}
	MemorySeg mem;
	mem.AddDefault(Data.latin1(), datalen);
	mem.AddDefault('\x00');
	datalen = b64_decode((unsigned char*)mem.BasePtr());
	ptrdiff_t offset=20;
	unsigned short KeyLen, FieldLen;
	char Key[256], Field[256];
	while(offset< datalen ) {
		mem.Extract(offset, &KeyLen, 2); offset+=2;
		KeyLen = ntohs(KeyLen);
		if(KeyLen > 32) {
			kdDebug (ICHAT_DEBUG_GLOBAL) << "Warning KeyLen("<<KeyLen <<") > 32: offset: " << offset << "datalen: " << datalen << endl;
			continue;
		}
		mem.Extract(offset, Key, KeyLen); offset+=KeyLen;
		Key[KeyLen]=0;
		mem.Extract(offset, &FieldLen, 2); offset+=2;
		FieldLen = ntohs(FieldLen)&0x7fff;
		if(FieldLen > 255) {
			kdDebug (ICHAT_DEBUG_GLOBAL) << "Warning DataLen > 255, key: " << Key << endl;
			continue;
		}
		mem.Extract(offset, Field, FieldLen); offset+=FieldLen;
		Field[FieldLen]=0;
		if(strcmp(Key, "phsh")==0) {
			char temp[64];
			memset(temp, 0, 64);
			int count=0;
			for(count=0;count<20;count++)
				snprintf(temp+count*2, 64-count*2, "%02hhx", (unsigned char)Field[count]);
			memcpy(Field, temp, 41);
			FieldLen=40;
		}
		Map[QString::fromUtf8(Key, KeyLen)] = QString::fromUtf8(Field, FieldLen);
	}
}


void iChatContact::slotRenameContact (const QString &oldName, const QString &newName)
{
	QString name = newName;

	kdDebug (ICHAT_DEBUG_GLOBAL) << k_funcinfo << "Renaming contact " << oldName << " to " << newName << endl;

	// if the name has been deleted, revert
	// to using the user ID instead
	if (name.isEmpty()) { kdDebug (ICHAT_DEBUG_GLOBAL)  << "Warning, name is empty" << endl;
		name = userId ();}

	setDisplayName (name);

}

void iChatContact::slotDeleteContact ()
{
	kdDebug (ICHAT_DEBUG_GLOBAL) << k_funcinfo << "Removing user " << userId () << endl;
	slotUpdatePresence(static_cast<iChatProtocol *>( protocol() ) ->iChatKOSOffline, QString::null );
	if(mNullQuery) {
		mNullQuery->Stop();
		mNullQuery = NULL;
	}
	static_cast<iChatAccount*>(account()) ->removeResolver(mResolve);
	mResolve = NULL;
	isLocalTyping=false;
	delete v4Addr;
	delete v6Addr;
	v4Addr = v6Addr = NULL;
	ChatStream.close();
	TXTFields.clear();

/*
	// unsubscribe
	if (!account()->isConnected ())
	{
		static_cast<iChatAccount *>(account())->errorConnectFirst ();
		return;
	}
*/
}

void iChatContact::slotTypingMsg(bool typing)
{
	if(isLocalTyping==typing) return;
	isLocalTyping=typing;
	Jabber::Message jabMessage(getv4Addy());
	jabMessage.setTyping(typing);
	if(ChatStream.isActive())
		ChatStream.sendPacket(jabMessage.toXml(&doc));
}


void iChatContact::km2jm (const KopeteMessage & km, Jabber::Message & jm)
{
	iChatContact *to = static_cast < iChatContact * >(km.to ().first ());
	const iChatContact *from = static_cast < const iChatContact * >(km.from ());

	if (!to || !from)
		return;

	Jabber::Message jabMessage (to->getv4Addy());
	jabMessage.setFrom(from->getv4Addy ());

	// if a resource has been selected for this contact,
	// send to this special resource - otherwise,
	// just send to the server and let the server decide
//	jabMessage.setTo (to->getv4Addy());

	//jabMessage.setFrom(from->userId();
	jabMessage.setSubject (km.subject ());
	jabMessage.setTimeStamp (km.timestamp ());

		// this message is not encrypted
		jabMessage.setBody (km.plainBody (), false);
		QString html = "<body ichattextcolor=\"" + km.fg().name() + "\"";
		if(km.bg() !=QColor("white"))
			html += " ichatballooncolor=\"" + km.bg().name() + "\"";
		html += ">\n <font color=\"" + km.fg().name() + "\" face=\"" + km.font().family() + "\" >" + km.escapedBody().replace("&nbsp;", " ") + "</font>\n</body>";
		// that replace for &nbsp; is because iChat can't parse &nbsp; for some strange reason. Maybe it's not part of XML, maybe it is. I don't know
		// All I know is that this makes it work :-P
//		html += ">\n" + km.escapedBody() + "</body>";
		jabMessage.setHTML( html );
//kdDebug (ICHAT_DEBUG_GLOBAL) << "XHTML of message to be sent:" << km.escapedBody() << endl;
//kdDebug (ICHAT_DEBUG_GLOBAL) << "Plain text of message to be sent:" << km.plainBody () << endl;

	// determine type of the widget and set message type accordingly
	//if (km.type() == KopeteMessage::Chat)
	if (messageManager->view ()->viewType () == KopeteMessage::Chat)
		jabMessage.setType ("chat");
	else
		jabMessage.setType ("normal");

	jm = jabMessage;

}

void iChatContact::slotReceivedMessage (const Jabber::Message & message)
{
	KopeteMessage::MessageType type;
	KopeteContactPtrList contactList;

	kdDebug (ICHAT_DEBUG_GLOBAL) << k_funcinfo << "Received Message Type:" << message.type () << endl;

	// check for errors first
	if (message.type () == "error")
	{
		// FIXME: unable to add i18n string here before release,
		// add warning dialog after 0.7
		return;
	}

	// determine message type
	if (message.type () == "chat")
		type = KopeteMessage::Chat;
	else
		type = KopeteMessage::Email;

	contactList.append (account()->myself ());

	// retrieve and reformat body
	QString body = message.body();
	QString xhtml = message.body(true);
	if(message.isTyping()) {
		manager()->receivedTypingMsg(this, true);
		return;
	}
	else
		manager()->receivedTypingMsg(this, false);

	if(xhtml.isEmpty()) {
	KopeteMessage newMessage (message.timeStamp (), this, contactList, body,
				  KopeteMessage::Inbound,
				  KopeteMessage::PlainText, type);
	// add it to the manager
	manager ()->appendMessage (newMessage);
	} else if(!body.isEmpty()) {
	KopeteMessage newMessage (message.timeStamp (), this, contactList, xhtml,
				  KopeteMessage::Inbound,
				  KopeteMessage::RichText, type);
	// add it to the manager
	manager ()->appendMessage (newMessage);
	}

	// convert Jabber::Message into KopeteMessage




}

void iChatContact::slotSendMessage (KopeteMessage & message)
{
	Jabber::Message jabberMessage;

	if (account()->isConnected ())
	{
		// convert the message
		km2jm (message, jabberMessage);
		if(!ChatStream.isActive()) {
			QObject::connect( &ChatStream, SIGNAL(connected()), this, SLOT(slotStreamConnect()) );
			QObject::connect( &ChatStream, SIGNAL(handshaken()), this, SLOT(slotStreamHandshake()));
			QObject::connect( &ChatStream, SIGNAL(error(const Jabber::StreamError& )), this, SLOT(slotStreamError(const Jabber::StreamError& )));
			QObject::connect(&ChatStream, SIGNAL(closeFinished()), this, SLOT(slotStreamClosed()));
			QObject::connect(&ChatStream, SIGNAL(receivePacket(const QDomElement& )), this, SLOT(slotStreamRecieve(const QDomElement&)));
			if(v4Addr) {
				kdDebug (ICHAT_DEBUG_GLOBAL) << "Reconnecting to...... " << v4Addr->pretty() <<endl;
				ChatStream.connectToHost(v4Addr->nodeName(), Port);
			} else if(v6Addr) {
				kdDebug (ICHAT_DEBUG_GLOBAL) << "Reconnecting to......"<< v6Addr->pretty() << endl;
				ChatStream.connectToHost(v6Addr->nodeName(), Port);
			}
			KopeteMessage ClosedWindowMessage ( this, account()->myself(), i18n("Reconnecting to %1...").arg(Name),
																		KopeteMessage::Internal,
																		KopeteMessage::PlainText, KopeteMessage::Chat);
			manager()->appendMessage(ClosedWindowMessage);

		}

		// send it
		// TODO SEND THE MESSAGE
//		static_cast<iChatAccount *>(account())->client()->sendMessage (jabberMessage);
		QDomElement xml = jabberMessage.toXml(&doc);
		QString data;
		QTextStream stream(&data, IO_WriteOnly) ;
		stream << xml;
		kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Sending message:" <<endl;
		kdDebug (ICHAT_DEBUG_GLOBAL) << data << endl;
		if(ChatStream.isActive())
			ChatStream.sendPacket(jabberMessage.toXml(&doc));

		isLocalTyping=false;
		// append the message to the manager
		manager ()->appendMessage (message);

		// tell the manager that we sent successfully
		manager ()->messageSucceeded ();
	}
	else
	{
		static_cast<iChatAccount *>(account())->errorConnectFirst ();
			kdDebug (ICHAT_DEBUG_GLOBAL) << "Apparently we are disconnected"<< endl;
		// FIXME: there is no messageFailed() yet,
		// but we need to stop the animation etc.
		manager ()->messageSucceeded ();
	}

}

#if 0
void iChatContact::slotUserInfo ()
{

	if (!static_cast<iChatAccount *>(account())->isConnected ())
	{
		static_cast<iChatAccount *>(account())->errorConnectFirst ();
		return;
	}

/*	Jabber::JT_VCard * task = new Jabber::JT_VCard (static_cast<JabberAccount *>(account())->client()->rootTask ());

	// signal to ourselves when the vCard data arrived
	QObject::connect (task, SIGNAL (finished ()), this, SLOT (slotGotVCard ()));

	task->get (userId ());

	task->go (true);
*/
}
#endif


void iChatContact::slotUserInfo ()
{
	kdDebug (ICHAT_DEBUG_GLOBAL) << "[iChatContact] Got vCard for user " << Name << ", displaying." << endl;

	dlgVCard = new dlgiChatVCard (qApp->mainWidget (), "dlgiChatVCard", this);

/*	if (mEditingVCard)
	{
		connect (dlgVCard, SIGNAL (saveAsXML (QDomElement &)), this, SLOT (slotSaveVCard (QDomElement &)));
		dlgVCard->setReadOnly (false);
		mEditingVCard = false;
	}*/
//	else
//		connect (dlgVCard, SIGNAL (updateNickname (const QString &)), this, SLOT (slotDoRenameContact (const QString &)));

	dlgVCard->show ();
	dlgVCard->raise ();

}

void iChatContact::serialize (QMap < QString, QString > &serializedData, QMap < QString, QString > & /* addressBookData */ )
{

	// Contact id and display name are already set for us, only add the rest
	serializedData["identityId"] = account()->accountId();

	serializedData["groups"] = QString::fromLatin1("Rendezvous");
//	serializedData["groups"] = rosterItem.groups ().join (QString::fromLatin1 (","));
}

#include "ichatcontact.moc"

// vim: set noet ts=4 sts=4 sw=4:
