 /*
  * ichatcontact.h  -  Base class for the Kopete iChat protocol contact
  *
  * Copyright (c) 2003 by Graham Dennis <u3952328@anu.edu.au>
  * Copyright (c) 2002-2003 by Till Gerken <till@tantalo.net>
  * Copyright (c) 2002 by Daniel Stone <dstone@kde.org>
  * 
  * Kopete    (c) by the Kopete developers  <kopete-devel@kde.org>
  * 
  * *************************************************************************
  * *                                                                       *
  * * This program is free software; you can redistribute it and/or modify  *
  * * it under the terms of the GNU General Public License as published by  *
  * * the Free Software Foundation; either version 2 of the License, or     *
  * * (at your option) any later version.                                   *
  * *                                                                       *
  * *************************************************************************
  */

#ifndef ICHATCONTACT_H
#define ICHATCONTACT_H

#include "ichataccount.h"
//#include "ichatresource.h"
#include "kopetecontact.h"
#include "rendezvous.h"
#include "stream.h"
#include "message.h"

class KAction;
class KListAction;
class KPopupMenu;
class KSelectAction;
class KInetSocketAddress;

class QImage;

//class dlgiChatRename;
class dlgiChatVCard;
class dlgVCard;
class iChatProtocol;
class iChatResource;
class iChatMessage;
class KopeteMessage;
class KopeteMessageManager;
class KopeteMetaContact;
class KopeteGroup;

class iChatContact:public KopeteContact
{
	Q_OBJECT

	 friend class iChatAccount;	/* Friends can touch each other's private parts. */

public:
	iChatContact (QString userid, RendezvousClient::ServiceInfo, iChatAccount * protocol, KopeteMetaContact * mc);
	iChatContact (QString userid, QString Nickname, iChatAccount *protocol, KopeteMetaContact *mc);	// TODO This needs to die.
	/********************************************************************
	 *
	 * KopeteContact reimplementation start
	 *
	 ********************************************************************/

	/**
	 * Return the user ID
	 */
	QString userId () const;

	/**
	 * Return the reason why we are away
	 */
	QString reason () const;

	/**
	 * Return if we are reachable (defaults to
	 * true because we can send on- and offline
	 */
	virtual bool isReachable ();

	/**
	 * Create custom context menu items for the contact
	 */
	virtual KActionCollection *customContextMenuActions ();


	/**
	 * Serialize contact
	 */
	virtual void serialize (QMap < QString, QString > &serializedData, QMap < QString, QString > &addressBookData);

	virtual KopeteMessageManager *manager (bool canCreate = false);


	QString getFirstName() const
	{
		if(!TXTFields.contains("1st"))
			return QString::null;
		return TXTFields["1st"];
	}

	QString getLastName() const
	{
		if(!TXTFields.contains("last"))
			return QString::null;
		return TXTFields["last"];
	}

	QString getEmail() const
	{
		if(!TXTFields.contains("email"))
			return QString::null;
		return TXTFields["email"];
	}

	QString getAIM() const
	{
		if(!TXTFields.contains("AIM"))
			return QString::null;
		return TXTFields["AIM"];
	}

	Q_UINT32 getPort() const
	{
		return Port;
	}

	bool getVideo() const
	{
		if(!TXTFields.contains("vc"))
			return false;
		return TXTFields["vc"].contains('V');
	}

	bool getAudio() const
	{
		if(!TXTFields.contains("vc"))
			return false;
		return TXTFields["vc"].contains('A');
	}

	QString getIdleStr() const;

	QString getStatusMsg() const
	{
		if(!TXTFields.contains("msg"))
			return QString::null;
		return TXTFields["msg"];
	}

	QPixmap& getBuddyImage()
	{
		return BuddyImage;
	}
	QString getv4Addy() const;
	QString getv6Addy() const;

	RendezvousClient::Resolve* Resolver() const { return mResolve; }
	RendezvousClient::NullQuery* NullQuery() const { return mNullQuery; }

public slots:

	/**
	 * Remove this contact from the roster
	 */
	virtual void slotDeleteContact ();

	/**
	 * Retrieve a vCard for the contact
	 */
	virtual void slotUserInfo ();

	/**
	 * vCard received from server for this contact
	 */
//	void slotGotVCard ();

	/**
	 * Update contact to new roster data
	 */
	void slotUpdateContact (const RendezvousClient::ServiceInfo);
	void slotUpdateContact (const RendezvousClient::ServiceInfo, const QMap<QString, QString>&);

	/**
	 * Update contact to a new status
	 */
	void slotUpdatePresence (const KopeteOnlineStatus & newStatus, const QString & reason);

	/**
	 * Received a message for this contact
	 */
	void slotReceivedMessage (const Jabber::Message & message);

	/**
	 * Sync Groups with server
	 */
//	virtual void syncGroups ();

	/**
	 * Incoming conversation -- the Account calls this
	 */
	void slotIncomingConversation(int sock, const QString& host, int port);

protected slots:

	/**
	 * Catch a dying message manager
	 */
	virtual void slotMessageManagerDeleted ();

private slots:

	/**
	 * Catch the rename dialog's results
	 */
	void slotRenameContact (const QString &oldName, const QString &newName);

	/**
	 * Edit a vCard for the contact.
	 */
//	void slotEditVCard ();

	/**
	 * Save this contact's vCard.
	 */
//	void slotSaveVCard (QDomElement &);

	/**
	 * Send a message to somebody
	 */
	void slotSendMessage (KopeteMessage & message);
	void slotTypingMsg(bool);

	void slotNullResult( RendezvousClient::NullQuery*, QString);

	void slotStreamConnect();
	void slotStreamHandshake();
	void slotStreamError(const Jabber::StreamError &);
	void slotStreamClosed();
	void slotStreamRecieve(const QDomElement&);


private:

	/**
	 * Convert KopeteMessage to Jabber::Message
	 */
	void km2jm (const KopeteMessage & km, Jabber::Message & jm);

	/**
	 * The metacontact that this contact belongs
	 * to.
	 */
	KopeteMetaContact *parentMetaContact;

	/**
	 * Current away reason and -type
	 */
	QString awayReason;

	/**
	 * Message manager in use to display a message
	 */
	KopeteMessageManager *messageManager;

//	KopeteMessageManager *m_manager;

	static void DecodeiChatTXTFields(const QStringList& , QMap<QString, QString>& );

	dlgiChatVCard *dlgVCard;
	QMap<QString, QString> TXTFields;
	RendezvousClient::Resolve *mResolve;
	RendezvousClient::NullQuery *mNullQuery;
	QString UserID;
	QString Name;
	KInetSocketAddress *v4Addr;
	KInetSocketAddress *v6Addr;
	QString Domain;
	QString FQDN;
	Jabber::Stream ChatStream;
	QDomDocument doc;
	Q_UINT32 Port;
	ulong IdleAt;
	QPixmap BuddyImage;
	Q_UINT8 phsh[20];
	bool isLocalTyping;
	enum ServiceName {
		_ichat,
		_presence,
		UnknownService
	};
	ServiceName Service;
};

#endif

// vim: set noet ts=4 sts=4 sw=4:
