/***************************************************************************
*   Copyright (C) 2003 by Graham Dennis                                   *
*   u3952328@anu.edu.au                                                   *
*                                                                         *
*   Permission is hereby granted, free of charge, to any person obtaining *
*   a copy of this software and associated documentation files (the       *
*   "Software"), to deal in the Software without restriction, including   *
*   without limitation the rights to use, copy, modify, merge, publish,   *
*   distribute, sublicense, and/or sell copies of the Software, and to    *
*   permit persons to whom the Software is furnished to do so, subject to *
*   the following conditions:                                             *
*                                                                         *
*   The above copyright notice and this permission notice shall be        *
*   included in all copies or substantial portions of the Software.       *
*                                                                         *
*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
*   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
*   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
*   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
*   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
*   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
*   OTHER DEALINGS IN THE SOFTWARE.                                       *
***************************************************************************/

#include "common.h"
#include "protocol.h"
#include <iostream>

#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>

using namespace Rendezvous;
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//								class IPCProtV1											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const char * IPCProtV1::SyncString = "Start";

//IPCProtV1::IPCProtV1() : IPCProtocolClass("1", "") {}
IPCProtV1::IPCProtV1(Instance*i) : IPCProtocol( 1, i ) {}

MemorySeg& IPCProtV1::EncodeMessage( Message& m, MemorySeg& Mem )
{
	uint32_t MessageSize = 0;
	uint16_t FieldCount = m.Data.size();
	Message::HashMapIterator i = m.Data.begin();
	Mem.AddDefault( SyncString, strlen( SyncString ) );	// 5 bytes for the syncstring
	Mem.Alloc( sizeof( MessageSize ) );			// 4 bytes at the start for MessageSize
	Mem.AddDefault( uint16_t(ntohs( FieldCount ) ) );	// 2 bytes for FieldCount
	assert( Mem.Size() == 6 + strlen(IPCProtV1::SyncString) );
	while ( FieldCount-- ) {
		Mem.AddDefault( uint16_t( htons( strlen( ( *i ).first ) ) ) );
		Mem.AddDefault( ( *i ).first, strlen( ( *i ).first ) );
		Mem.AddDefault( uint32_t( htonl( ( *i ).second.first )  ) );
		Mem.AddDefault( ( *i ).second.second, ( *i ).second.first );
		++i;
	}
	MessageSize = htonl(Mem.Size());
	Mem.Write(strlen(IPCProtV1::SyncString), &MessageSize, sizeof(uint32_t));
	return Mem;
}

int IPCProtV1::DecodeMessages( IPC* ipc )
{
	char * key;
	uint32_t MessageSize;
	uint16_t FieldCount;
	uint16_t keylen;
	uint32_t fieldlen;
	size_t offset;
	size_t BufSize;
	MemorySeg *Buffer;
	do {
		offset = 0;
		Buffer = BufferPtr( ipc );
		if ( Buffer->BasePtr() == NULL ) return 0;
		BufSize = Buffer->Size();
		if ( BufSize < strlen( SyncString ) ) return 0;
		if ( memcmp( Buffer->BasePtr(), SyncString, strlen( SyncString ) ) != 0 ) {
			cout << "No SyncString!!" << endl;
			if ( Resync( ipc ) == -1 )
				return -1;
		}
		offset += strlen( SyncString );
		Buffer->Extract( offset, &MessageSize, sizeof( uint32_t ) );
		MessageSize = ntohl( MessageSize );
		if ( BufSize < MessageSize ) return 0;
		offset += sizeof( uint32_t );
		Buffer->Extract( offset, &FieldCount, sizeof( uint16_t ) );
		FieldCount = ntohs( FieldCount );
		offset += sizeof( uint16_t );
		Message *m = new Message;

		while ( FieldCount-- ) {
			Buffer->Extract( offset, &keylen, sizeof( uint16_t ) );
			keylen = ntohs( keylen );
			offset += sizeof( uint16_t );
			if ( offset + keylen > MessageSize ) return -1;
			key = (char*) malloc( keylen + 1);
			Buffer->Extract( offset, key, keylen );
			key[ keylen ] = 0;
			offset += keylen;
			Buffer->Extract( offset, &fieldlen, sizeof( uint32_t ) );
			fieldlen = ntohl( fieldlen );
			offset += sizeof( uint32_t );
			if ( offset + fieldlen > MessageSize ) return -1;
			m->AddData( key, Buffer->Offset( offset ), fieldlen );
			offset += fieldlen;
			free(key);
		}
		MSGList( ipc ).push_back( m );
		Buffer->FreeFirst( MessageSize );
	} while ( 1 );

	return 0;


}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//							class UDSocketIPC											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const char* const UDSocketIPC::SocketPath = "/tmp/.mdns_socket";


UDSocketIPC::UDSocketIPC( Instance*i, int newsocket, struct sockaddr_un* newremote ) : SocketIPC( i, newsocket )
{
	if ( newremote )
		memcpy( &remote, newremote, sizeof( remote ) );
	else
		memset(&remote, 0, sizeof(remote));
}

UDSocketIPC::~UDSocketIPC()
{
	Disconnect();
}

int UDSocketIPC::Connect()
{
	int len;
	int sock;

	if ( ( sock = socket( AF_UNIX, SOCK_STREAM, 0 ))  == -1 ) {
		perror( "socket" );
		return -1;
	}
	printf("Obtained socket: %i\n", sock);

	printf( "Trying to connect...\n" );

	remote.sun_family = AF_UNIX;
	strcpy( remote.sun_path, UDSocketIPC::SocketPath );
	len = strlen( remote.sun_path ) + sizeof( remote.sun_family );
	if ( connect( sock, ( struct sockaddr * ) & remote, len ) == -1 ) {
		perror( "connect" );
		return -1;
	}

	printf( "Connected on UD socket %i.\n", sock );
	Status = Negotiating;
	useSocket(sock);
	IPC::NegotiateProtocol();

	return 0;


}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//							class TCPSocketIPC											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

const char* const TCPSocketIPC::DefaultHost = "ursulad37.anu.edu.au";

TCPSocketIPC::TCPSocketIPC( Instance *i, int newsocket, struct sockaddr_in* newremote ) : SocketIPC( i, newsocket )
{
	if ( newremote )
		memcpy( &remote, newremote, sizeof( remote ) );
	else
		memset(&remote, 0, sizeof(remote));
}

TCPSocketIPC::TCPSocketIPC(Instance* i, const char*const newHost, unsigned int newPort) :
		 SocketIPC(i), Host(newHost), Port(newPort) {}

TCPSocketIPC::~TCPSocketIPC()
{
	Disconnect();
}

int TCPSocketIPC::Connect()
{
	int sock;
	struct hostent *he;

	if((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		perror("socket");
		return -1;
	}

	MyOptions = CSocket::Blocking;
	printf("Obtained socket: %i\n", sock);

	printf( "Trying to connect...\n" );

	if((he=gethostbyname(Host.c_str())) == NULL)
	{
		perror("gethostbyname");
		return -1;
	}
	remote.sin_family = AF_INET;				/* host byte order	*/
	remote.sin_port = htons(Port);		/* short, network byte order	*/
	remote.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&(remote.sin_zero), '\0', 8);		/* zero the rest of the struct	*/

	if(connect(sock, (struct sockaddr *)&remote, sizeof(struct sockaddr)) == -1)
	{
		perror("connect");
		return -1;
	}
	printf( "Connected on TCP socket %i.\n", sock );
	Status = Negotiating;
	useSocket(sock);
	setOptions(CSocket::NonBlocking);
	IPC::NegotiateProtocol();

	return 0;


}

