/*
 * stream.cpp - handles a Jabber XML stream
 * Copyright (C) 2001, 2002  Justin Karneges
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define QSSL_STATIC

#include"stream.h"

#include<qlibrary.h>
#include<qregexp.h>
#include<qptrqueue.h>
#include<qsocket.h>
#include<qtimer.h>
#include<qfileinfo.h>

#include"xmlfilter.h"

using namespace Jabber;

//----------------------------------------------------------------------------
// StreamError
//----------------------------------------------------------------------------

//! \brief Creates StreamError object
//!
//! \param _type - type of error
//! \param _string - error message for that type
//! \param _isWarning - if this is a warning or an error
StreamError::StreamError(int _type, const QString &_string, bool _isWarning)
{
	v_type = _type;
	v_string = _string;
	v_isWarning = _isWarning;
}

//! \brief Returns if this is a warning or error
//!
//! \return true: if warning\n
//!   false: if error
bool StreamError::isWarning() const
{
	return v_isWarning;
}

//! \brief Returns the error/warrning type.
int StreamError::type() const
{
	return v_type;
}

//! \brief Returns the error/warning information.
const QString & StreamError::details() const
{
	return v_string;
}

//! \brief creates a warning/error string for you.
//!
//! This function is here for your convenience. This will return a neatly
//! formated error/warrning message.
QString StreamError::toString() const
{
	QString str;
	if(isWarning())
		str += Stream::tr("Warning");
	else
		str += Stream::tr("Error");
	str += ": ";
//	enum { DNS, Refused, Timeout, Socket, Disconnected, Handshake, SSL, Proxy, Unknown };
	switch(type()) {
		case Refused:
			str += Stream::tr("Connection Refused"); break;
		case Timeout:
			str += Stream::tr("Connection Timeout"); break;
		case Socket:
			str += Stream::tr("Socket"); break;
		case DNS:
			str += Stream::tr("DNS"); break;
		case Disconnected:
			str += Stream::tr("Disconnected"); break;
		case Handshake:
			str += Stream::tr("Handshake"); break;
		case Unknown:
		default:
			break;
	}
	if(!details().isEmpty()) {
		str += ": ";
		str += details();
	}

	return str;
}


//----------------------------------------------------------------------------
// Stream
//----------------------------------------------------------------------------
//! \if _hide_doc_
class Stream::StreamPrivate
{
public:
	StreamPrivate() {}

	XmlFilter xml;
	QSocket *sock;

	QTimer *noopTimer;
	bool isActive, isConnected, isHandshaken, closing;
	QString host;
	int port;

	bool http_inHeader;

	StreamError err;
	int noop_time;

	QString id;
	QCString Buffer;

	QPtrQueue<QDomElement> in;

};
//! \endif

//! \brief Create Stream object.
//!
//! \param QObject - pass on to QObject
Stream::Stream(QObject *par)
:QObject(par)
{
	d = new StreamPrivate;

	d->closing = false;
	d->sock = 0;
	d->isActive = d->isConnected = d->isHandshaken = false;

	d->noop_time = 0;
	d->noopTimer = new QTimer;
	connect(d->noopTimer, SIGNAL(timeout()), SLOT(doNoop()));

	d->in.setAutoDelete(true);

	connect(&d->xml, SIGNAL(packetReady(const QDomElement &)), SLOT(xml_packetReady(const QDomElement &)));
	connect(&d->xml, SIGNAL(handshake(bool, const QString &)), SLOT(xml_handshake(bool, const QString &)));

}

//! \brief Destroys Stream object
Stream::~Stream()
{
	close();

	delete d->noopTimer;

}

//! \brief Tells you if the connection is busy.
//!
//! \return true: is busy\n
//!   false: nothings going on
bool Stream::isActive() const
{
	return d->isActive;
}

//! \brief Tells you if the connection is establised.
bool Stream::isConnected() const
{
	return d->isConnected;
}

//! \brief Tells you if the handshakeing process has been completed.
bool Stream::isHandshaken() const
{
	return d->isHandshaken;
}

//! \brief Closes socket connection.
//!
//! This function will disconnect and clean this object out.
void Stream::close()
{
	if(!d->isActive)
		return;

	if(d->closing)
		return;
	d->Buffer.resize(0);

	//printf("stream: closing...\n");

	if(d->sock) {
		if(d->sock->state() == QSocket::Connected) {
			if(d->isHandshaken)
				sendString("</stream:stream>\n");

			d->closing = true;

			int bytesLeft = d->sock->bytesToWrite();
			//printf("stream: bytesToWrite: %d\n", bytesLeft);

			d->sock->close();

			if(bytesLeft == 0)
				QTimer::singleShot(0, this, SLOT(afterClose()));
			else {
				//printf("stream: socket shutting down...\n");
			}
		}
		else {
			cleanup();
		}
	}
	else
		QTimer::singleShot(0, this, SLOT(afterClose()));
}

void Stream::afterClose()
{
	//printf("stream: socket closed.\n");
	cleanup();
	closeFinished();
}

void Stream::cleanup()
{
	//printf("stream: cleanup()\n");
	if(d->sock) {
		delete d->sock;
		d->sock = 0;

		if(d->isConnected)
			d->xml.reset();

	}

	d->noopTimer->stop();
	d->isActive = d->isConnected = d->isHandshaken = false;
	d->closing = false;
}

//! \brief Sets the host to connect to.
//!
//! \param _host - Sets the IP of the host you want to connect to.
//! \param _port - Sets the port to connect.
void Stream::connectToHost(const QString &_host, int _port)
{
	if(d->isActive)
		return;

	d->host = _host;

	d->port = _port;

	d->isActive = true;

	d->sock = new QSocket;
	connect(d->sock, SIGNAL(connectionClosed()), SLOT(sock_disconnected()));
	connect(d->sock, SIGNAL(error(int)),  SLOT(sock_error(int)));
	connect(d->sock, SIGNAL(bytesWritten(int)), SLOT(sock_bytesWritten(int)));
	connect(d->sock, SIGNAL(delayedCloseFinished()), SLOT(sock_delayedCloseFinished()));

	connect(d->sock, SIGNAL(connected()), SLOT(sock_connected()));
	connect(d->sock, SIGNAL(readyRead()), SLOT(sock_readyRead()));
	d->sock->connectToHost(_host, _port);

}

void Stream::setSocket( int _sock, const QString& _host, int _port)
{
	if(d->isActive)
		return;

	d->host = _host;

	d->port = _port;

	d->isActive = true;

	d->sock = new QSocket;
	connect(d->sock, SIGNAL(connectionClosed()), SLOT(sock_disconnected()));
	connect(d->sock, SIGNAL(error(int)),  SLOT(sock_error(int)));
	connect(d->sock, SIGNAL(bytesWritten(int)), SLOT(sock_bytesWritten(int)));
	connect(d->sock, SIGNAL(delayedCloseFinished()), SLOT(sock_delayedCloseFinished()));

	connect(d->sock, SIGNAL(connected()), SLOT(sock_connected()));
	connect(d->sock, SIGNAL(readyRead()), SLOT(sock_readyRead()));
	d->sock->setSocket(_sock);
	sock_connected();
}

QCString Stream::base64Encode( const QCString &s )
{
	int i;
	int len = s.length();
	char tbl[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
	int a, b, c;

	QCString p = QCString( "" );
	for( i = 0; i < len; i += 3 ) {
		a = (s[i] & 3) << 4;
		if(i + 1 < len) {
			a += s[i + 1] >> 4;
			b = (s[i + 1] & 0xF) << 2;
			if(i + 2 < len) {
				b += s[i + 2] >> 6;
				c = s[i + 2] & 0x3F;
			}
			else
				c = 64;
		}
		else
			b = c = 64;

		p += tbl[s[i] >> 2];
		p += tbl[a];
		p += tbl[b];
		p += tbl[c];
	}
	return p;
}

//! \brief Return stream id.
QString Stream::id() const
{
	return d->id;
}

//! \brief Set ping interval.
//!
//! This section will set the interval to send new lines to keep the conenction
//! alive.
void Stream::setNoopTime(int mills)
{
	d->noop_time = mills;

	if(!d->isHandshaken)
		return;

	if(d->noop_time == 0) {
		d->noopTimer->stop();
		return;
	}

	d->noopTimer->start(d->noop_time);
}

//! \brief Send xml to host.
void Stream::sendPacket(const QDomElement &e)
{
	sendString(elemToString(e));
}

//! \brief send String to host.
void Stream::sendString(const QCString &str)
{
	if(d->isConnected)
		d->sock->writeBlock(str, str.length());
	else if(d->isActive)
		d->Buffer +=str;
}

void Stream::sock_connected()
{
	d->isConnected = true;
	d->xml.begin();
	emit connected();
	startHandshake();
}

void Stream::sock_disconnected()
{
	if(d->closing)
		return;

	d->err = StreamError(StreamError::Disconnected);

	// process error later
	QTimer::singleShot(0, this, SLOT(delayedProcessError()));
}

void Stream::sock_readyRead()
{
	int size;
	QByteArray buf;

	size = d->sock->bytesAvailable();
	buf.resize(size);
	d->sock->readBlock(buf.data(), size);

	processIncomingData(buf);
}

void Stream::sock_error(int x)
{
	if(x == QSocket::ErrConnectionRefused)
		d->err = StreamError(StreamError::Refused);
	else if(x == QSocket::ErrHostNotFound)
		d->err = StreamError(StreamError::DNS);
	else if(x == QSocket::ErrSocketRead)
		d->err = StreamError(StreamError::Socket);
	else
		d->err = StreamError(StreamError::Timeout);

	// process error later
	QTimer::singleShot(0, this, SLOT(delayedProcessError()));
}

void Stream::sock_bytesWritten(int)
{
	//printf("stream: wrote %d. left=%d\n", (int)x, (int)d->sock->bytesToWrite());
}

void Stream::sock_delayedCloseFinished()
{
	//printf("stream: sock::delayedCloseFinished()\n");
	afterClose();
}

void Stream::startHandshake()
{
	// Start the handshake
	QCString str;
	str.sprintf("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n<stream:stream to=\"%s\" xmlns=\"jabber:client\" xmlns:stream=\"http://etherx.jabber.org/streams\">\n", encodeXML(d->host).data());
	sendString(str);
}

void Stream::processIncomingData(const QByteArray &buf)
{
	d->xml.putIncomingXmlData(buf);
}

void Stream::delayedProcessError()
{
	if(!d->err.isWarning())
		close();
	error(d->err);
}

void Stream::delayedProcessReceived()
{
	// process chunks
	while(!d->in.isEmpty()) {
		QDomElement *e = d->in.dequeue();
		receivePacket(*e);
		delete e;
	}
}

void Stream::delayedProcessHandShake()
{
	d->isHandshaken = true;

	setNoopTime(d->noop_time);
	sendString(d->Buffer);
	d->Buffer.resize(0);
	handshaken();
}

void Stream::doNoop()
{
	if(d->isHandshaken)
		sendString("\n");
}

void Stream::xml_packetReady(const QDomElement &e)
{
	d->in.enqueue(new QDomElement(e));
	QTimer::singleShot(0, this, SLOT(delayedProcessReceived()));
}

void Stream::xml_handshake(bool ok, const QString &id)
{
	if(!ok) {
		d->err = StreamError(StreamError::Handshake);

		// process error later
		QTimer::singleShot(0, this, SLOT(delayedProcessError()));
		return;
	}

	d->id = id;

	// process handshake later
	QTimer::singleShot(0, this, SLOT(delayedProcessHandShake()));
}

//! \brief Set the XML string.
QCString Stream::encodeXML(const QString &_str)
{
	QString str = _str;

	str.replace(QRegExp("&"), "&amp;");
	str.replace(QRegExp("<"), "&lt;");
	str.replace(QRegExp(">"), "&gt;");
	str.replace(QRegExp("\""), "&quot;");
	str.replace(QRegExp("'"), "&apos;");

	return str.utf8();
}

//! \brief Convert QDomElemnt to QCString.
QCString Stream::elemToString(const QDomElement &e)
{
	QString out;
	QTextStream ts(&out, IO_WriteOnly);
	e.save(ts, 0);
	return out.utf8();
}

#include "stream.moc"
