//
//
// C++ Interface: rendezvous
//
// Description:
//
//
// Author: Graham Dennis <u3952328@anu.edu.au>, (C) 2003
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef __RENDEZVOUS_H
#define __RENDEZVOUS_H

#include "common.h"
#include "ipc.h"
#include "protocol.h"
#include "rez.h" 
//#include "rezclient.h"

#include <qobject.h>
#include <qsocket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qmap.h>

using namespace Rendezvous;
namespace RendezvousClient
{
	class Client;
	class Action;
	class Service;
	class Browse;
	class Query;
	class Resolve;
	struct ServiceInfo
	{
		//		unsigned int ID;
		RendezvousClient::Resolve* Resolver;
		QString FQDN;
		QString Name;
		QString Type;
		QString Domain;
		QString Address;
		bool isAddrV4;
		unsigned int Port;
		QStringList TXTFields;
	};

	struct BrowseInfo
	{
		//		unsigned int ID;
		RendezvousClient::Browse* Browser;
		QString Name;
		QString Type;
		QString Domain;
		QString FQDN;
		bool Active;
	};

	class Client : public QObject, public IPC
	{
		Q_OBJECT
	private:
		Instance __Instance;
		QSocket	sockmgr;
		struct sockaddr_un remote;
		CSocket::ConnectionStatus Status;
		fd_set IDList;
		unsigned int AvailID;
		QMap<unsigned int, Action*> ActionsMap;

		int ReceiveAll();
		int SendMemSeg( MemorySeg& );

		static int	Callback	( Message*, IPC* );
		unsigned int getID();
		void freeID( unsigned int );
		int IDused( unsigned int );
		int Stop( unsigned int, const char* );
		unsigned int RegisterAction( Action* );
		int DeregisterAction( unsigned int ID, Action* );

		friend class Action;
	protected:
		CSocket::ConnectionStatus setStatus( CSocket::ConnectionStatus );
		CSocket::ConnectionStatus getStatus() { return Status; }
		int IPCDisconnect() { return Disconnect(); }
		int IPCReceiveAll() { return ReceiveAll(); }

	public:
		Client();
		~Client();

		int Connect();
		int Disconnect();

		CSocket::ConnectionStatus ConnStatus();
		int StopAll();

	signals:
		void sigConnected();
		void sigDisconnected();
		void sigSocketError( int );

	public slots:
		void slotConnect();
		void slotDisconnect();

	private slots:
		void slotIncoming();
		void slotSocketError( int );
		void slotSocketDisconnected();
	};


	class Action
	{
	private:
		unsigned int mID;
		bool stopping;
	protected:
		Client * const mClient;
		unsigned int getID() { return mID; }
	public:
		Action( Client* c ) : mClient( c ) { mID = mClient->RegisterAction( this ); stopping=false;}
		virtual ~Action() { if ( mID ) mClient->DeregisterAction( mID, this ); }

		virtual int Callback( Message* ) = 0;
		virtual int Stop() = 0;
		int StopAction( const char* s );
		int SendMessage( Message& m )
		{
			assert( mClient );
			if ( mID == 0 ) mID = mClient->RegisterAction( this );
			m.AddTranslate( "ID", mID );
			return mClient->SendMessage( m );
		}
bool isRunning() { return mID ? true : false; }

		Client* getClient() const { return mClient; }

		static void ConvertTXT2Map( const QStringList&, QMap<QString, QString>& );
		static void ConvertMap2TXT( const QMap<QString, QString>&, QStringList& );
		static void ConvertString2TXT( const QString&, QStringList& );
		static int TranslateTXT( Message*, const char* prefix, QStringList& );

	};

	class Query : public QObject, public Action
	{
		Q_OBJECT
	public:
		Query( Client* c ) : Action( c ) {}
		~Query() { Stop(); }
		int Start( QString Query, QString QueryType, QString Domain = QString::null );
		int Stop() { return StopAction( "StopQuery" ); }
		int Callback( Message* );
	signals:
		void sigQueryResult( RendezvousClient::Query*, Message* );
		void sigQueryResultExpire( RendezvousClient::Query*, Message* );
	};

	class Browse : public QObject, public Action
	{
		Q_OBJECT
	public:
		Browse( Client* c ) : Action( c ) {}
		~Browse() { Stop(); }
		int Start( QString Type, QString Domain = QString::null );
		int Stop() { return StopAction( "StopBrowse" ); }
		int Callback( Message* );
	signals:
		void sigBrowseResult( RendezvousClient::BrowseInfo );
	};

	class Resolve : public QObject, public Action
	{
		Q_OBJECT
	public:
		Resolve( Client* c ) : Action( c ) {}
		~Resolve() { Stop(); }
		int Start( QString Name, QString Type, QString Domain = QString::null );
		int Start( QString FQDN );
		int Stop() { return StopAction( "StopResolve" ); }
		int Callback( Message* );
	signals:
		void sigServiceResolved( RendezvousClient::ServiceInfo );
	};

	struct DNSType {
		enum {
	    DNSType_A = 1,  			//  1 Address
	    DNSType_NS,  			//  2 Name Server
	    DNSType_MD,  			//  3 Mail Destination
	    DNSType_MF,  			//  4 Mail Forwarder
	    DNSType_CNAME,  			//  5 Canonical Name
	    DNSType_SOA,  			//  6 Start of Authority
	    DNSType_MB,  			//  7 Mailbox
	    DNSType_MG,  			//  8 Mail Group
	    DNSType_MR,  			//  9 Mail Rename
	    DNSType_NULL,  			// 10 NULL RR
	    DNSType_WKS,  			// 11 Well-known-service
	    DNSType_PTR,  			// 12 Domain name pointer
	    DNSType_HINFO,  			// 13 Host information
	    DNSType_MINFO,  			// 14 Mailbox information
	    DNSType_MX,  			// 15 Mail Exchanger
	    DNSType_TXT,  			// 16 Arbitrary text string

	    DNSType_AAAA = 28,  		// 28 IPv6 address
	    DNSType_SRV = 33  		// 33 Service record
		};
		int Type;
		DNSType() { Type = 0; }
		DNSType(const int t) { Type = t; }
		DNSType(const DNSType& old) { Type = old.Type; }
	};

	class ExtraRecord
	{
	private:
		Service* _service;
		const int RR_ID;
		const DNSType _type;
		MemorySeg _mem;
		unsigned int _ttl;
	public:
		ExtraRecord( Service *, int RR_ID, const DNSType Type );
		~ExtraRecord();

		int Stop();
		int Restart();

		void setData( QByteArray&, unsigned int ttl = 0 );
		void setData( QCString&, unsigned int ttl = 0 );
		void setData( void* data, size_t len, unsigned int ttl = 0 );
		DNSType type();
		int id();
	};

	class Service : public QObject, public Action
	{
		Q_OBJECT
	private:
		friend class ExtraRecord;

		int RegisterService( QString Name, QString Type, unsigned int Port, QStringList&, QString Domain = QString::null );
		int UpdateTXTRecord( QStringList&, unsigned int ttl = 60 );
		QPtrList<ExtraRecord> ExtraRecordList;
	public:
		Service( Client* c ) : Action( c ) {}
		int Start( QString Name, QString Type, unsigned int Port, QString TXTRecord, QString Domain = QString::null );
		int Start( QString Name, QString Type, unsigned int Port, QMap<QString, QString>& TXTMap,
		           QString Domain = QString::null );
		int Start( QString Name, QString Type, unsigned int Port, QStringList, QString TXTRecord, QString Domain = QString::null );
		int Start( QString Name, QString Type, unsigned int Port, QStringList, QMap<QString, QString>& TXTMap,
		           QString Domain = QString::null );
		~Service();
		int Stop();
		int Callback( Message* );

		int UpdateTXTRecord( QString TXTRecord, unsigned int ttl = 60 );
		int UpdateTXTRecord( QMap<QString, QString>& TXTMap, unsigned int ttl = 60 );
		int UpdateTXTRecord( QStringList, QString TXTRecord, unsigned int ttl = 60 );
		int UpdateTXTRecord( QStringList, QMap<QString, QString>& TXTMap, unsigned int ttl = 60 );
		// FIXME: We cannot handle raw data as a TXT record. But that is fine for iChat use :-)

		ExtraRecord* newExtraRecord( DNSType Type );
		bool removeExtraRecord( ExtraRecord* );
		/*		int AddExtraRecordToService( QString data, QString Type, unsigned int ttl = 0);
				int UpdateExtraRecord( int RR_ID, QByteArray& _ba, QString Type,*/
	signals:
		void sigRegisterServiceCallback( RendezvousClient::Service*, QString status );
	};

	class NullQuery : public QObject, public Action
	{
		Q_OBJECT
	public:
		NullQuery( Client* c ) : Action( c ) {}
		~NullQuery() { Stop(); }
		int Start( QString FQDN );
		int Stop() { return StopAction( "StopQuery" ); }
		int Callback( Message* );
	signals:
		void sigNullResult( RendezvousClient::NullQuery*, QString );
		void sigNullExpire( RendezvousClient::NullQuery*, QString );
	};

};

#endif
