/***************************************************************************
*   Copyright (C) 2003 by Graham Dennis                                   *
*   u3952328@anu.edu.au                                                   *
*                                                                         *
*   Permission is hereby granted, free of charge, to any person obtaining *
*   a copy of this software and associated documentation files (the       *
*   "Software"), to deal in the Software without restriction, including   *
*   without limitation the rights to use, copy, modify, merge, publish,   *
*   distribute, sublicense, and/or sell copies of the Software, and to    *
*   permit persons to whom the Software is furnished to do so, subject to *
*   the following conditions:                                             *
*                                                                         *
*   The above copyright notice and this permission notice shall be        *
*   included in all copies or substantial portions of the Software.       *
*                                                                         *
*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
*   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
*   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
*   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
*   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
*   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
*   OTHER DEALINGS IN THE SOFTWARE.                                       *
***************************************************************************/

#ifndef __COMMON_H
#define __COMMON_H

#include <list>
#include <typeinfo>

#ifndef __WIN32
#include <netinet/in.h>
#include <sys/time.h>
#else
#include <windows.h>
#define NO_UNIX_SOCKETS
#endif

#define debug_locate()   // dlocate(__FILE__, __FUNCTION__, __LINE__)

using namespace std;
#ifdef __GNUC__
#if __GNUC__ < 3
	#include <hash_map.h>
	// inherit globals
#else
	#include <ext/hash_map>
	#if __GNUC_MINOR__ == 0
//          namespace Sgi = std;               // GCC 3.0
	#else
	using namespace __gnu_cxx;       // GCC 3.1 and later
	#endif
#endif
#else      // ...  there are other compilers, right?
#include <hash_map.h>
//      namespace Sgi = std;
#endif

namespace Rendezvous
{

	void dlocate( const char* const file, const char* const function, int line );

	template <class T> inline T& MAX( T& left, T& right ) { return left < right ? right : left; }
	template <class T> inline T& MIN( T& left, T& right ) { return left < right ? left : right; }

	class MemorySeg;
	class FDSet;
	class FDGroup;
	class Message;
	class CSocket;
	class IPC;
	class SocketIPC;



	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//								Memory Classes											//
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


	class MemorySeg
	{
		void *Base;
		size_t BufSize;
		size_t BufUsed;
		size_t MinSizeInc;

		static const size_t DefaultMinSizeInc = 100;

	public:

		MemorySeg( MemorySeg& );
		MemorySeg( size_t BufSize = 0, size_t MinSizeInc = MemorySeg::DefaultMinSizeInc );
		~MemorySeg();

		void* Alloc( size_t size );
		template <class T> T* AddDefault( T* Data, size_t nbytes ) { debug_locate(); return reinterpret_cast<T*>( memcpy( Alloc( nbytes ), Data, nbytes ) ); }
		// Should this cast be changed?
		template <class T> T* AddDefault( T Data ) { debug_locate(); return reinterpret_cast<T*>( memcpy( Alloc( sizeof( T ) ), &Data, sizeof( T ) ) ); }

		int Extract( size_t offset, void *buf, size_t nbytes );
		int Write( size_t offset, void* buf, size_t nbytes );
		void *Offset( size_t offset );
		int ZeroMemory();
		size_t Size();
		size_t Allocated();
		void *BasePtr();
		void Free();
		void FreeFirst( size_t nbytes );
		operator char*();
		operator void*();
		const MemorySeg& operator=( const MemorySeg& );
	};


	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//								Communication Classes									//
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	class FDSet
	{
		fd_set fdSet;
	public:
		int fdmax;


		FDSet();
		FDSet( FDSet & );

		FDSet operator= ( FDSet& old );
		int IsSet( const int fd );
		void Set( const int fd );
		void Unset( const int fd );
		void Zero();
		void Add( fd_set*, int fdmax );
		fd_set *SetPtr();
	};


	class FDGroup
	{
	public:
		FDSet Read;
		FDSet Write;
		FDSet Except;
		struct timeval timeout;
		static const int DefaultTimeout = 1;
		FDGroup();

		FDGroup operator= ( FDGroup& old );
		int MaxFD();
	};


	struct eqstr
	{
		bool operator() ( const char* s1, const char* s2 ) const
		{
			return strcmp( s1, s2 ) == 0;
		}
	};


	class Message
	{
	protected:
		template <class T> void Add_internal( const char* Key, const char*format, T& value )
		{
			const unsigned int StringLen = 30;
			char *string = new char [ StringLen ];
			snprintf( string, StringLen, format, value );
			Add( Key, string );
			delete [] string;
		}
		void Addu8( const char* Key, const uint8_t& );
		void Addi8( const char* Key, const int8_t& );
		void Addu16( const char* Key, const uint16_t& );
		void Addi16( const char* Key, const int16_t& );
		void Addu32( const char* Key, const uint32_t& );
		void Addi32( const char* Key, const int32_t& );
		void Addu64( const char* Key, const uint64_t& );
		void Addi64( const char* Key, const int64_t& );

	public:
		typedef hash_map<const char*, pair<uint32_t, void*>, hash<const char*>, eqstr > HashMap;
		typedef Message::HashMap::iterator HashMapIterator;
		typedef	list<Message*>	MessageList;
		typedef	Message::MessageList::iterator MessageListIterator;
		enum Options {
		    AppendNullTerminator,
		    NoNullTerminator,
		    Duplicate,
		    DoNotDuplicate
		};

		Message::HashMap Data;

		Message();
		Message( Message* );
		Message(Message& msg);
		virtual ~Message();


		void Add( const char* Key, const char* string , Message::Options = AppendNullTerminator );
		void AddData( const char* Key, const void *Buffer, size_t BufSize , Message::Options = Duplicate );
		void Add( const char* Key, MemorySeg& MemSeg );
		void Add( const char* Key, MemorySeg* MemSeg );

		void Add( const char* Key, const unsigned char&value );
		void Add( const char* Key, const signed char& value );
		void Add( const char* Key, const unsigned int &value );
		void Add( const char* Key, const signed int &value );
		void Add( const char* Key, const unsigned long &value );
		void Add( const char* Key, const signed long &value );

		template <class T>
		void AddTranslate( const char* Key, const T& value )
		{
			if ( sizeof( T ) == 1 ) Addu8( Key, uint8_t( value ) );
			else if ( sizeof( T ) == 2 ) Addu16( Key, uint16_t( value ) );
			else if ( sizeof( T ) == 4 ) Addu32( Key, uint32_t( value ) );
			else {
				//			cout << "ERROR: No Add to deal with type \"" << typeid(value).name() << '"'<<endl;
//				cout << "ERROR: No Add to deal with type" << endl;
				assert( 0 );
			}
		}

		void AddKey( const char* NewKey, const char* OldKey, Message* m );
		void AddKey( const char*Key, Message *m );

		void Function( const char* func );
		void Callback( const char* type );
		const char* Function();
		const char* CallbackType();

	};

};

#endif

