/***************************************************************************
*   Copyright (C) 2003 by Graham Dennis                                   *
*   u3952328@anu.edu.au                                                   *
*                                                                         *
*   Permission is hereby granted, free of charge, to any person obtaining *
*   a copy of this software and associated documentation files (the       *
*   "Software"), to deal in the Software without restriction, including   *
*   without limitation the rights to use, copy, modify, merge, publish,   *
*   distribute, sublicense, and/or sell copies of the Software, and to    *
*   permit persons to whom the Software is furnished to do so, subject to *
*   the following conditions:                                             *
*                                                                         *
*   The above copyright notice and this permission notice shall be        *
*   included in all copies or substantial portions of the Software.       *
*                                                                         *
*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
*   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
*   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
*   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
*   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
*   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
*   OTHER DEALINGS IN THE SOFTWARE.                                       *
***************************************************************************/


#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <iostream>

#include "common.h"
#include "ipc.h"
#include "rez.h"

using namespace Rendezvous;


namespace Rendezvous {
class InstancePrivate
{
	public:
		IPCProtV1	ipcv1;
		MSGProtV1	msgv1;

		InstancePrivate(Instance* i) : ipcv1(i), msgv1(i) {}
		~InstancePrivate() {}
};
}

Instance::Instance()
{
	d = new InstancePrivate(this);
}

Instance::~Instance()
{
	IPCProtocols.clear();
	MSGProtocols.clear();
	delete d;
	d=NULL;
}

int Instance::AddHandler(const unsigned int ver, const char* name, MSGProtocol::MessageHandler handler)
{
	if(MSGProtocols.count(ver)==0) return -1;
	MSGProtocols[ver]->MHMap[name] = handler;
	return 0;
}

int Instance::RemoveHandler(const unsigned int ver, const char* name)
{
	if(MSGProtocols.count(ver)==0) return -1;
	if(MSGProtocols[ver]->MHMap.count(name)==0) return -1;
	MSGProtocols[ver]->MHMap.erase(name);
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//							class MSGProtocolClass										//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

MSGProtocol::MSGProtocol( const unsigned int v, Instance* i) : mInstance(i), Version( v )
{
	mInstance->MSGProtocols[Version]=this;
}

MSGProtocol::~MSGProtocol()
{
	if(mInstance->MSGProtocols.count(Version)==1)
		mInstance->MSGProtocols.erase((unsigned int)Version);
}

Message::MessageList& MSGProtocol::MSGList( IPC* ipc ) { return ipc->IncomingMessages; }


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//							class IPCProtocolClass										//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

IPCProtocol::IPCProtocol( const unsigned int v , Instance* i) : mInstance(i), Version( v )
{
	mInstance->IPCProtocols[Version]=this;
}

IPCProtocol::~IPCProtocol()
{
	if(mInstance->IPCProtocols.count(Version)==1)
		mInstance->IPCProtocols.erase(Version);
}

int IPCProtocol::Resync( IPC* ipc )
{
	ipc->Disconnect();
	return -1;
}

MemorySeg& IPCProtocol::Buffer( IPC* ipc ) { return ipc->Buffer; }
MemorySeg* IPCProtocol::BufferPtr( IPC* ipc ) { return & ipc->Buffer; }

Message::MessageList& IPCProtocol::MSGList( IPC* ipc ) { return ipc->IncomingMessages; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//								class CSocket											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

FDGroup CSocket::fds;
int CSocket::Running = 1;
CSocket::Options CSocket::DefaultOptions = CSocket::NonBlocking;
CSocket::ProcessFcnList CSocket::ProcessFunctions;

CSocket::CSocket( int sock )
{
	Socket = -1;
	MyOptions = CSocket::DefaultOptions;
	useSocket( sock );
}
/*
CSocket::CSocket( CSocket *oldsock )
{
	Socket = oldsock->Socket;
	MyOptions = oldsock->MyOptions;
}*/

CSocket::~CSocket()
{
	Disconnect();
}

int CSocket::useSocket( int sock )
{
	if ( Socket != -1 || sock==-1) return -1;
	Socket = sock;
	CSocket::fds.Read.Set( sock );

	if ( MyOptions == NonBlocking ) {
		if ( fcntl( Socket, F_SETFL, O_NONBLOCK ) == -1 ) {
			perror( "fcntl" );
			return -1;
		}
	}
	CSocket::ProcessFunctions[ Socket ] = this;
	return 0;
}

int CSocket::setOptions(CSocket::Options opt)
{
	if(Socket==-1) return -1;
	if (opt == NonBlocking) {
		if ( fcntl( Socket, F_SETFL, O_NONBLOCK ) == -1 ) {
			perror( "fcntl" );
			return -1;
		}
	} else if (opt == Blocking) {
		if(fcntl(Socket, F_SETFL, 0) == -1) {
			perror( "fcntl" );
			return -1;
		}
	}
	return 0;

}

int CSocket::Disconnect()
{
	if(Socket==-1) return -1;
	CSocket::ProcessFunctions.erase(Socket);
	if ( close( Socket ) == -1 ) {
		perror( "close" );
		Socket = -1;
		return -1;
	}
	cout << "Disconnecting socket " << Socket << endl;
	CSocket::fds.Read.Unset( Socket );
	Socket = -1;
	return 0;
}

int CSocket::getSocket() { return Socket; }

int CSocket::GetFDSet( int &nfds, fd_set* read_fds, struct timeval* timeout)
{
	(void)timeout;	// Unused parameter
	for (int i=CSocket::fds.MaxFD()-1; i>=0;i--)
	{
		if(CSocket::fds.Read.IsSet(i))
			FD_SET(i, read_fds);
	}
	if(nfds < CSocket::fds.MaxFD() )
		nfds = CSocket::fds.MaxFD();

	return 0;
}

int CSocket::ProcessFDSet( fd_set* read_fds)
{
	CSocket::PFLiterator i = CSocket::ProcessFunctions.begin();
	int sock;
	CSocket* sockptr;

	for ( int k = CSocket::ProcessFunctions.size(); k > 0;k-- ) {
		sock = (*i).first;
		sockptr = (*i).second;
		++i;
		if ( FD_ISSET(sock, read_fds) )
			sockptr->ProcessIncoming();
		if ( sockptr->Socket == -1 ) {
			cout << "Deleting socket: " << sock << endl;
			delete sockptr;
			if(CSocket::ProcessFunctions.count( sock ) ==1) {
				cout << "Warning, the socket has not been removed from ProcessFunctions...This shouldn't happen..." << endl;
				CSocket::ProcessFunctions.erase( sock );
			}
		}
	}
	return 0;
}

int CSocket::SelectAndRun()
{
	int ret;
	FDGroup LocalFDs = CSocket::fds;
	CSocket::Running = 1;
	int sock;
	CSocket* sockptr;

	if ( ( ret = select( LocalFDs.MaxFD(), LocalFDs.Read.SetPtr(), LocalFDs.Write.SetPtr(),
	                     LocalFDs.Except.SetPtr(), &LocalFDs.timeout ) ) == -1 ) {
		perror( "select" );
		CSocket::Running=0;
		return -1;
	}

	CSocket::PFLiterator i = CSocket::ProcessFunctions.begin();
	for ( int k = CSocket::ProcessFunctions.size(); k > 0;k-- ) {
		sock = (*i).first;
		sockptr = (*i).second;
		++i;
		if ( LocalFDs.Read.IsSet( sock ) )
			sockptr->ProcessIncoming();
		else if ( LocalFDs.Write.IsSet( sock) )
			sockptr->ProcessIncoming();
		else if ( LocalFDs.Except.IsSet( sock) )
			sockptr->ProcessIncoming();
		if ( sockptr->Socket == -1 ) {
			cout << "Deleting socket: " << sock << endl;
			delete sockptr;
			if(CSocket::ProcessFunctions.count( sock ) ==1) {
				cout << "Warning, the socket has not been removed from ProcessFunctions...This shouldn't happen..." << endl;
				CSocket::ProcessFunctions.erase( sock );
			}
		}
	}

	return ret;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//									class IPC											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

IPC::IPC(Instance* i) : mInstance(i)
{
	IPCprot=NULL;
	MSGprot=NULL;
}

IPC::~IPC()
{
	Buffer.Free();
}

CSocket::ConnectionStatus IPC::NegotiateProtocol()
{
	if(mInstance->IPCProtocols.count(1)==0 || mInstance->MSGProtocols.count(1)==0) {
		IPCDisconnect();
		if(mInstance->IPCProtocols.count(1)==0)
			cout << "No IPC Prot v1" << endl;
		if(mInstance->MSGProtocols.count(1)==0)
			cout << "No msg prot v1" << endl;
		cout << "we are disconnecting" << endl;
		return CSocket::Disconnected;
	}
	IPCprot = mInstance->IPCProtocols[1];
	MSGprot = mInstance->MSGProtocols[1];
	return setStatus(CSocket::Connected);
}

int IPC::IPCProcessIncoming()
{
	if ( getStatus() == CSocket::Negotiating ) cout << "Status is Negotiating" << endl;
if ( getStatus() == CSocket::Disconnected ) { cout << "Disconnected" << endl; return -1; }
	if ( IPCReceiveAll() == -1 ) { return -1; }
	switch ( getStatus() ) {
		case CSocket::Negotiating:

			if ( NegotiateProtocol() != CSocket::Connected )
				return 0;

		case CSocket::Connected:
		if ( !IPCprot || !MSGprot ) { IPCDisconnect(); cout << "Although connected, we don't have a MSG or IPC protocol" << endl; return -1;}
			if ( IPCprot->DecodeMessages( this ) == -1 )
				return -1;
			return MSGprot->ProcessMessages( this );
		default:
			return -1;
	}
	return 0;
}

int IPC::SendMessage( Message& msg )
{
	if ( !IPCprot || !MSGprot ) {
		if ( IPCprot == NULL ) cout << "IPC Protocol is NULL" << endl;
		if ( MSGprot == NULL ) cout << "MSGProtocol is NULL" << endl;
		cout << "No IPC or MSG protocols" << endl;
		return -1;
	}
	MemorySeg Mem;
	return SendMemSeg( IPCprot->EncodeMessage( msg, Mem ) );
}

int IPC::SendError( Message* m, const char*error)
{
	if(error==NULL) return -1;
	Message msg;
	msg.Function("Error");
	msg.Add("Error", error);
	msg.AddKey( "InFunction", "Function", m);
/*	if(id)
		msg.Add("ID", id);
*/	return SendMessage(msg);
}

int IPC::SendError( const char*error)
{
	if(error==NULL) return -1;
	Message msg;
	msg.Function("Error");
	msg.Add("Error", error);
/*	if(id)
		msg.Add("ID", id);
*/	return SendMessage(msg);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//								class SocketIPC											//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

SocketIPC::SocketIPC( Instance* i, int newsocket) : CSocket( newsocket ), IPC(i)
{
	if ( newsocket == -1 ) Status = Disconnected;
	else	Status = Negotiating;
}

SocketIPC::~SocketIPC()
{
	Disconnect();
}

/*CSocket::ConnectionStatus SocketIPC::ConnStatus() { return Status; }
CSocket::ConnectionStatus SocketIPC::setConnStatus(CSocket::ConnectionStatus con) { return Status = con; }*/

int SocketIPC::SendAll( char *buf, int *len, int flags )
{
	if ( Status == Disconnected ) return -1;
	int total = 0;			/* how many bytes we've sent		*/
	int bytesleft = *len;	/* how many we have left to send	*/
	int n=-1;
	while ( total < *len ) {
		n = send( getSocket(), buf + total, bytesleft, flags );
		if ( n == -1 ) {
			break;
		}
		total += n;
		bytesleft -= n;

	}
	*len = total;	/* return number actually sent here	*/

	return n == -1 ? -1 : 0;
}

int SocketIPC::ReceiveAll( size_t MaxReceive, int flags )
{
	if ( Status == Disconnected ) return -1;
	char *tmpbuf = new char [ MaxReceive ];
	int nbytes;
	int totalbytes = 0;
	int running = 1;
	while ( running ) {
		if ( ( nbytes = recv( getSocket(), tmpbuf, MaxReceive, flags ) ) == -1 ) {
			switch ( errno ) {
				case EBADF:   				// Bad file descriptor
				case ECONNREFUSED:   	// Connection refused
				case ENOTCONN:   		// Not connected
				case ENOTSOCK:   		// socket isn't a socket
				case EINVAL:   			// Generic failure
					cout << "Error before disconnect" << endl;
					Disconnect();
				case EINTR:   				// interrupted... apart from doing this again, I don't know what to do
				case EFAULT:   			// Memory addressing error with tmpbuf... same as above.
					cout << "Error after disconnect" << endl;
					perror( "recv" );
					delete [] tmpbuf;
					return -1;

				case EAGAIN:   			// Non-blocking socket. All is OK.
					cout << "EAGAIN: totalbytes = " << totalbytes << endl;
					perror( "recv" );
					if ( totalbytes == 0 ) {
						delete [] tmpbuf;
						return 0;
					} else running = 0;
					break;

				default:
					Disconnect();
					cout << "default error" << endl;
					perror( "recv" );
					delete [] tmpbuf;
					return -1;
			}
		}
		if ( nbytes == 0 ) {
			Disconnect();
			delete [] tmpbuf;
			return -1;
		}
		if ( nbytes != -1 ) {
			Buffer.AddDefault( tmpbuf, nbytes );
			totalbytes += nbytes;
		}
		if ( nbytes < (signed int) MaxReceive ) running = 0;
	}
	cout << "Received " << totalbytes << " bytes" << endl;
	delete [] tmpbuf;
	return nbytes;
}

int SocketIPC::SendMemSeg( MemorySeg& m )
{
	if ( Status == Disconnected ) { cout << "Apparently we are disconnected" << endl; return -1; }
	int len = m.Size();
	return SendAll( m, &len );
}

int SocketIPC::Disconnect()
{
	Buffer.Free();
	Message::MessageListIterator i = IncomingMessages.begin();
	int k = IncomingMessages.size();
	Message* temp;
	for (;k>0;k--)
	{
		temp = *i;
		++i;
		delete temp;
	}
	CSocket::Disconnect();
	Status = Disconnected;
	return 0;
}
