/***************************************************************************
*   Copyright (C) 2003 by Graham Dennis                                   *
*   u3952328@anu.edu.au                                                   *
*                                                                         *
*   Permission is hereby granted, free of charge, to any person obtaining *
*   a copy of this software and associated documentation files (the       *
*   "Software"), to deal in the Software without restriction, including   *
*   without limitation the rights to use, copy, modify, merge, publish,   *
*   distribute, sublicense, and/or sell copies of the Software, and to    *
*   permit persons to whom the Software is furnished to do so, subject to *
*   the following conditions:                                             *
*                                                                         *
*   The above copyright notice and this permission notice shall be        *
*   included in all copies or substantial portions of the Software.       *
*                                                                         *
*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,       *
*   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF    *
*   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.*
*   IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR     *
*   OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, *
*   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR *
*   OTHER DEALINGS IN THE SOFTWARE.                                       *
***************************************************************************/

#ifndef __IPC_H
#define __IPC_H

#include "common.h"

#include <netinet/in.h>
#include <sys/un.h>

namespace Rendezvous
{

	class MSGProtocol;
	class IPCProtocol;
	class CSocket;
	class IPC;
	class SocketIPC;
	class Instance;

	class MSGProtocol
	{
	private:
		Instance* const mInstance;
	public:
		const unsigned int Version;
		typedef int ( *MessageHandler ) ( Message*, IPC* );
		typedef hash_map<const char*, MSGProtocol::MessageHandler, hash<const char*>, eqstr>
									MessageHandlerMap;
		MSGProtocol::MessageHandlerMap MHMap;

		MSGProtocol( const unsigned int vers, Instance*);
		virtual ~MSGProtocol();

		virtual int ProcessMessages( IPC* ) = 0;
		Message::MessageList& MSGList( IPC* );
	};


	class IPCProtocol
	{
	private:
		Instance* const mInstance;
	public:
		const unsigned int Version;

		IPCProtocol( const unsigned int Version, Instance* );
		virtual ~IPCProtocol();

		virtual MemorySeg&	EncodeMessage( Message&, MemorySeg& ) = 0;
		virtual int	DecodeMessages( IPC* ) = 0;
		virtual int	Resync( IPC* );
		MemorySeg& Buffer( IPC* );
		MemorySeg* BufferPtr( IPC* );
		Message::MessageList& MSGList( IPC* );

	};

	class CSocket
	{
	private:
		int Socket;
	public:
		enum Options {
		    Blocking,
		    NonBlocking
		};
		enum ConnectionStatus {
		    Disconnected,
		    Negotiating,
		    Listening,
		    Connected
		};

		typedef hash_map<int, CSocket*> ProcessFcnList;
		typedef ProcessFcnList::iterator PFLiterator;

		static FDGroup fds;
		static int Running;
		static ProcessFcnList ProcessFunctions;
		static CSocket::Options DefaultOptions;

		static int GetFDSet( int &nfds, fd_set* read_fds, struct timeval* timeout);
		static int ProcessFDSet( fd_set* read_fds);
		static int SelectAndRun();

		CSocket::Options MyOptions;
		CSocket::ConnectionStatus Status;

		CSocket( int =-1);
//		CSocket( CSocket* );
		virtual ~CSocket();

		virtual int ProcessIncoming() = 0;
		virtual int Disconnect();


		int getSocket();
		int useSocket( int );

		int setOptions( CSocket::Options );

	};

	class IPC
	{
	protected:
		IPCProtocol*	IPCprot;
		MSGProtocol*	MSGprot;
		Message::MessageList	IncomingMessages;
		MemorySeg	Buffer;
		Instance* const mInstance;


		virtual int SendMemSeg( MemorySeg& ) =0;
		virtual int IPCDisconnect()=0;
		virtual int IPCReceiveAll()=0;
		virtual CSocket::ConnectionStatus setStatus(CSocket::ConnectionStatus)=0;
		virtual CSocket::ConnectionStatus getStatus()=0;

		friend class IPCProtocol;
		friend class MSGProtocol;
	public:

		IPC(Instance*);
		virtual ~IPC();

		CSocket::ConnectionStatus NegotiateProtocol();

		virtual int SendMessage( Message & );

		virtual int IPCProcessIncoming();
		virtual int Disconnect() =0;
		virtual int SendError( Message* m, const char*error );
		virtual int SendError( const char* error);

	};

	class SocketIPC : public CSocket, public IPC
	{

	protected:
		virtual int SendAll( char *buf, int *len, int flags = 0 );
		virtual int ReceiveAll( size_t maxreceive = 1000, int flags = 0 );
		virtual int SendMemSeg( MemorySeg & );
		int IPCReceiveAll() { return ReceiveAll(); }
		int IPCDisconnect() { return Disconnect(); }

	public:
		static FDGroup fds;

		SocketIPC( Instance*, int newsocket = -1 );
		//	SocketIPC(SocketIPC*);
		virtual ~SocketIPC();

		CSocket::ConnectionStatus getStatus() { return Status; }
		CSocket::ConnectionStatus setStatus(CSocket::ConnectionStatus s) { return Status=s;}

		virtual int Connect() = 0;
		virtual int ProcessIncoming() { return IPCProcessIncoming(); }
		virtual int Disconnect();
		int IPCgetSocket() { return getSocket(); }

	};
	class InstancePrivate;
	class Instance
	{
		public:
		hash_map<unsigned int, IPCProtocol*>	IPCProtocols;
		hash_map<unsigned int, MSGProtocol*>	MSGProtocols;
		int AddHandler(const unsigned int, const char*, MSGProtocol::MessageHandler);
		int RemoveHandler(const unsigned int, const char*);

		Instance();
		~Instance();

		private:
		InstancePrivate *d;
 };


};

#endif


