/*
 * message.cpp - message handling classes
 * Copyright (C) 2001, 2002  Justin Karneges
 *                           Akito Nozaki
 *                           Hideaki Omuro
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

//! \class Jabber::Message message.h
//! \brief handles all incomming/outgoing messages.
//!
//! Message member is a very handy tool that will help you process the
//! message information faster. This function will do all the XML processing
//! to convert it into Message object format.
//!
//! The next example will show you how to recive and send the message.
//! \code
//! #include <message.h>
//!
//! ...
//! client = new Jabber::Client;
//! ...
//!
//! ...
//! // Setup where the message is going to go when received
//! connect(client, SIGNAL(messageReceived(const Message &)), SLOT(clientMessageReceived(const Message &)));
//! ...
//!
//! ...
//! clientMessageReceived(const Jabber::Message &m)
//! {
//! 	// Once the message has been recieved you can print the information out.
//! 	printf("message received!\n");
//! 	printf("From: %s\n", m.from().full().latin1());
//! 	printf("Subject: %s\n", m.subject().latin1());
//! 	printf("Body: %s\n", m.body().latin1());
//! 	for(Jabber::UrlList::Iterator it = m.urlList().begin(); it != m.urlList().end(); ++it)
//!     	printf("URL: %s {%s}\n", (*it).url().latin1(), (*it).desc().latin1());
//!
//! 	// Well for example... lets forward this message to 2 people =)
//! 	// Its probably easier if i just set the "to" to my jid but i'll do it the hard way.
//! 	Jabber::Message fwd(Jid("anpluto@orbit.tenpura.org"));
//! 	// Set the subject
//! 	fwd.setSubject(m.subject());
//! 	// Since we want to know who it came from
//! 	fwd.setBody("From: m.from().full\n\n" + m.body());
//!     // Set the url list
//! 	fwd.setUrlList(m.urlList());
//!
//! 	// Ok, finally. Lets send it
//! 	client->sendMessage(fwd);
//!
//! 	// OK, lets send it off to one more person
//! 	fwd.setTo(Jid("justin@orbit.tenpura.org"));
//! 	client->sendMessage(fwd);
//! }
//! \endcode
//!
//! Pretty long example but this should show everything you need to
//! recieve/send message to your friends.
//!
//! Don't use this exact code to test. I don't want millions of message
//! forwarded to me... =P
//!
//! \sa Jabber::Client
#include"message.h"

#include<qdatetime.h>
#include<qdom.h>

#include"xmlcommon.h"

#include<qtextstream.h>
#include <iostream>

/*
static QByteArray nodeToArray(const QDomNode &e)
{
	QString out;
	QTextStream ts(&out, IO_WriteOnly);
	e.save(ts, 1);
	QCString xmlToEnc = out.utf8();
	//printf("{%s}\n", out.data());
	int len = xmlToEnc.length();
	QByteArray b(len);
	memcpy(b.data(), xmlToEnc.data(), len);
	return b;
}*/

using namespace Jabber;

//! \if _hide_doc_
class Message::MessagePrivate
{
public:
	MessagePrivate() {}

	QString to, from;

	QString type;
	QString subject;
	QString bodyPlain, bodyHTML;
	QDateTime timeStamp;
	int errorCode;
	QString errorString;
	bool isTyping;

	QString invite;

	QDomElement out;
	bool flag;
};
//! \endif

//----------------------------------------------------------------------------
// Message
//----------------------------------------------------------------------------
//! \brief Constructs Message with given Jid information.
//!
//! This function will construct a Message container.
//! \param to - specify reciver (default: empty string)
Message::Message(const QString &to)
{
	d = new MessagePrivate;
	d->to = to;
	d->flag = false;
	d->errorCode = -1;
	d->isTyping=false;
}

//! \brief Constructs a copy of Message object
//!
//! Overloaded constructor which will constructs a exact copy of the Message
//! object that was passed to the constructor.
//! \param from - Message object you want to copy
Message::Message(const Message &from)
{
	d = 0;
	*this = from;
}

//! \brief Required for internel use.
Message & Message::operator=(const Message &from)
{
	delete d;
	d = new MessagePrivate;
	*d = *from.d;

	return *this;
}

//! \brief Destroy Message object.
Message::~Message()
{
	delete d;
	d = 0;
}

//! \brief Return receiver's Jid information.
QString Message::to() const
{
	return d->to;
}

//! \brief Return sender's Jid information.
QString Message::from() const
{
	return d->from;
}

//! \brief Return body information.
//!
//! This function will return a plain text or the Richtext version if it
//! it exists.
//! \param rich - Returns XHTML if true and plain text if false. (default: false)
//! \note Richtext is in Qt's richtext format and not in xhtml.
QString Message::body(bool html) const
{
	if(html)
		return d->bodyHTML;
	else
		return d->bodyPlain;
}

//! \brief Return subject information.
QString Message::subject() const
{
	return d->subject;
}

//! \brief Return type information
QString Message::type() const
{
	return d->type;
}

//! \brief Return time information
QDateTime Message::timeStamp() const
{
	return d->timeStamp;
}

//! \brief Return error string
QString Message::errorString() const
{
	return d->errorString;
}

void Message::setError(int code, const QString &str)
{
	d->errorCode = code;
	d->errorString = str;
}

QString Message::invite() const
{
	return d->invite;
}

bool Message::isTyping() const
{
	return d->isTyping;
}

void Message::setTyping(bool _isTyping)
{
	d->isTyping = _isTyping;
}

//! \brief Set receivers information
//!
//! \param to - Receivers Jabber id
void Message::setTo(const QString &to)
{
	d->to = to;
	d->flag = false;
}

void Message::setFrom(const QString &from)
{
	d->from = from;
	d->flag = false;
}

//! \brief Set subject
//!
//! \param subject - Subject information
void Message::setSubject(const QString &subject)
{
	d->subject = subject;
	d->flag = false;
}

//! \brief Set body
//!
//! \param body - body information
//! \param rich - set richtext if true and set plaintext if false.
//! \note Richtext support will be implemented in the future... Sorry.
void Message::setBody(const QString &body, bool)
{
	d->bodyPlain = body;
	d->flag = false;
}

void Message::setHTML(const QString &html)
{
	d->bodyHTML = html;
	d->flag = false;
}


//! \brief Set Type of message
//!
//! \param type - type of message your going to send
void Message::setType(const QString &type)
{
	d->type = type;
	d->flag = false;
}

void Message::setTimeStamp(const QDateTime &ts)
{
	d->timeStamp = ts;
}

void Message::setInvite(const QString &i)
{
	d->invite = i;
}

//! \brief Set XML dom element
//!
//! \param out - xml information
void Message::setAsXml(const QDomElement &out)
{
	d->out = out;
	d->flag = true;
}

//! \brief convert QDomDocument to QDomElement
//!
//! Create message xml ready to be sent out.
//! \param doc -
QDomElement Message::toXml(QDomDocument *doc)
{
	if(d->flag)
		return d->out;

	QDomElement message = doc->createElement("message");
	if(!d->to.isEmpty())
		message.setAttribute("to", d->to);
	if(!d->from.isEmpty())
		message.setAttribute("from", d->from);
	if(!d->type.isEmpty())
		message.setAttribute("type", d->type);

	if(!d->subject.isEmpty())
		message.appendChild(textTag(doc, "subject", d->subject));

	if(!d->bodyPlain.isEmpty())
		message.appendChild(textTag(doc, "body", d->bodyPlain));
	else {
		QDomElement body = doc->createElement("body");
		message.appendChild(body);
	}
	if(!d->bodyHTML.isEmpty()) {
/*		std::cout << "bodyHTML isn't empty...making modifications" << std::endl;
		std::cout << "It is: " << d->bodyHTML << std::endl;*/
/*		QDomElement html = doc->createElement("html");
		html.setAttribute(QString::fromLatin1("xmlns"), QString::fromLatin1("http://www.w3.org/1999/xhtml"));
		QDomElement body = doc->createElement("body");
		body.setAttribute(QString::fromLatin1("ichattextcolor"), QString::fromLatin1("#000000"));
	*/	QDomDocument msgdoc;
		if(msgdoc.setContent(d->bodyHTML,false)==true)
		{
/*			body.appendChild(msgdoc.documentElement());
			html.appendChild(body);*/
			QDomElement html = doc->createElement("html");
			html.setAttribute(QString::fromLatin1("xmlns"), QString::fromLatin1("http://www.w3.org/1999/xhtml"));
			html.appendChild(msgdoc.documentElement());
			QString output;
			QTextStream mystream(&output, IO_WriteOnly);
			mystream << html;
/*			std::cout << "HTML of our message: " << output << std::endl;*/
			message.appendChild(html);

		} /*else
			std::cout << "Apparently bodyHTML couldn't be parsed as XML" << std::endl;*/
	}

	if(d->errorCode >= 0) {
		QDomElement e = textTag(doc, "error", d->errorString);
		e.setAttribute("code", QString::number(d->errorCode));
		message.appendChild(e);
	}

	QDomElement x = doc->createElement("x");
	x.setAttribute(QString::fromLatin1("xmlns"), QString::fromLatin1("jabber:x:event"));
	// The only time we don't send <composing/> is when we aren't typing and both bodyPlain and bodyHTML are empty
	if(!( !d->isTyping && d->bodyPlain.isEmpty() && d->bodyHTML.isEmpty())) {
		QDomElement composing = doc->createElement("composing");
		x.appendChild(composing);
	}
	// The only time we don't send <id/> is when we are sending an actual message. ie typing==false, and one or other of the body fields are non-empty
	if(!(!d->isTyping && ( !d->bodyPlain.isEmpty() || !d->bodyHTML.isEmpty() ))) {
		QDomElement id = doc->createElement("id");
		x.appendChild(id);
	}
	message.appendChild(x);

/*
	for(QValueList<Url>::Iterator it = d->urlList.begin(); it != d->urlList.end(); ++it) {
		QDomElement x = doc->createElement("x");
		x.setAttribute("xmlns", "jabber:x:oob");
		x.appendChild(textTag(doc, "url", (*it).url()));
		if(!(*it).desc().isEmpty())
			x.appendChild(textTag(doc, "desc", (*it).desc()));
		message.appendChild(x);

	}
*/
	if(!d->invite.isEmpty()) {
		QDomElement e = doc->createElement("x");
		e.setAttribute("xmlns", "jabber:x:conference");
		e.setAttribute("jid", d->invite);
		message.appendChild(e);
	}

	d->out = message;
	return d->out;
}

//! \brief Conver XML encoded message into Message object.
//!
//! This function will return false if the XML chunk is not a message chunk.
//! \param e - XML encoded chunk
//! \param timeZoneOffset - If you need to set any time zone offset.
bool Message::fromXml(const QDomElement &e, int timeZoneOffset)
{
	if(e.tagName() != "message")
		return false;

	QDomElement tag;
	bool found;
	d->timeStamp = QDateTime::currentDateTime();
	QString errorString;

	d->to = e.attribute("to");
	d->from = e.attribute("from");
	d->type = e.attribute("type");

	d->bodyPlain = "";
	d->bodyHTML = "";
	tag = findSubTag(e, "body", &found);
	if(found)
		d->bodyPlain = tagContent(tag);

	tag = findSubTag(e, "subject", &found);
	if(found)
		d->subject = tagContent(tag);

	tag = findSubTag(e, "error", &found);
	if(found)
		d->errorString = tagContent(tag);
	tag = findSubTag(e, "html", &found);
	if(found) {
/*		std::cout << "HTML tag found" << std::endl;*/
		QTextStream stream(&d->bodyHTML, IO_WriteOnly) ;

		for(QDomNode n = tag.firstChild(); !n.isNull(); n = n.nextSibling()) {
			QDomElement i = n.toElement();
			if(i.isNull())
				continue;
			if(i.tagName() == "body") {
				for(QDomNode p = n.firstChild(); !p.isNull(); p = p.nextSibling()) {
					stream << p;
				}
			}
		}

//		stream << tag;
/*		std::cout << "HTMLContent:" << d->bodyHTML <<std::endl;*/
	} /*else
		std::cout << "HTML tag not found" << std::endl;*/

	for(QDomNode n = e.firstChild(); !n.isNull(); n = n.nextSibling()) {
		QDomElement i = n.toElement();
		if(i.isNull())
			continue;

		if(i.tagName() == "x" && i.attribute("xmlns") == "jabber:x:delay") {
			if(i.hasAttribute("stamp")) {
				if(stamp2TS(i.attribute("stamp"), &d->timeStamp)) {
					d->timeStamp = d->timeStamp.addSecs(timeZoneOffset * 3600);
				}
			}
		}

		if(i.tagName() == "x" && i.attribute("xmlns") == "jabber:x:event") {
			QDomElement tag;
			bool found;
			tag = findSubTag(i, "id", &found);
			if(found)
				d->isTyping = true;
		}

/*		if(i.tagName() == "x" && i.attribute("xmlns") == "jabber:x:oob") {
			QDomElement tag;
			bool found;
			Url u;

			tag = findSubTag(i, "url", &found);
			if(found)
				u.setUrl(tagContent(tag));
			tag = findSubTag(i, "desc", &found);
			if(found)
				u.setDesc(tagContent(tag));

			d->urlList += u;
		}*/

		if(i.tagName() == "x" && i.attribute("xmlns") == "jabber:x:conference") {
			d->invite = i.attribute("jid");
		}
	}

	return true;
}
