 /*
  *   rendezvousprotocol.cpp  -  Base class for the Kopete Rendezvous protocol
  *
  * Copyright (c) 2003 by Graham Dennis <u3952328@anu.edu.au>
  * Copyright (c) 2002-2003 by Till Gerken <till@tantalo.net>
  * Copyright (c) 2002 by Daniel Stone <dstone@kde.org>
  *
  *  Kopete   (c) by the Kopete developers  <kopete-devel@kde.org>
  *
  * *************************************************************************
  * *                                                                       *
  * * This program is free software; you can redistribute it and/or modify  *
  * * it under the terms of the GNU General Public License as published by  *
  * * the Free Software Foundation; either version 2 of the License, or     *
  * * (at your option) any later version.                                   *
  * *                                                                       *
  * *************************************************************************
  */

#include <kdebug.h>
#include <kgenericfactory.h>
#include <kconfig.h>
#include <kmessagebox.h>
#include <kiconloader.h>
#include <kpopupmenu.h>
#include <kstandarddirs.h>
#include <klineeditdlg.h>

#include <qapplication.h>
#include <qcursor.h>
#include <qmap.h>
#include <qtimer.h>
#include <qpixmap.h>
#include <qstringlist.h>

//#include "client.h"
#include "stream.h"
//#include "tasks.h"
//#include "types.h"
//#include "vcard.h"

#include <sys/utsname.h>

#include "../kopete/kopete.h"
#include "kopetecontact.h"
#include "kopetecontactlist.h"
#include "kopetemetacontact.h"
#include "kopetemessagemanager.h"
#include "kopeteaway.h"
#include "kopeteprotocol.h"
#include "kopeteplugin.h"
#include "addcontactpage.h"
#include "rendezvouscontact.h"
#include "rendezvousaddcontactpage.h"
#include "rendezvousprotocol.h"
#include "rendezvousaccount.h"
#include "kopeteaccountmanager.h"
#include "rendezvouseditaccountwidget.h"
#include "kopetecommandhandler.h"


RendezvousProtocol *RendezvousProtocol::protocolInstance = 0;

//K_EXPORT_COMPONENT_FACTORY (kopete_rendezvous, KGenericFactory < RendezvousProtocol >);
typedef KGenericFactory<RendezvousProtocol> RendezvousProtocolFactory;

K_EXPORT_COMPONENT_FACTORY( kopete_rendezvous, RendezvousProtocolFactory( "kopete_rendezvous" )  )

RendezvousProtocol::RendezvousProtocol (QObject * parent, const char *name, const QStringList &) : KopeteProtocol( RendezvousProtocolFactory::instance(), parent, name ),
	RendezvousKOSOnline(KopeteOnlineStatus::Online,         25, this, 0, QString::null,       i18n ("O&nline"), i18n ("Online")),
	RendezvousKOSIdle(KopeteOnlineStatus::Away,             25, this, 2, "rendezvous_away",       i18n ("&Idle"), i18n ("Idle")),
	RendezvousKOSDND(KopeteOnlineStatus::Away,              15, this, 4, "rendezvous_na",         i18n ("&Away"), i18n ("Away")),
	RendezvousKOSOffline(KopeteOnlineStatus::Offline,       20, this, 5, QString::null,       i18n ("O&ffline"), i18n ("Offline")),
	RendezvousKOSConnecting(KopeteOnlineStatus::Connecting,  2, this, 7, "rendezvous_connecting", i18n ("FIXME: You should not see this"), i18n("Connecting"))
{

	kdDebug (RENDEZVOUS_DEBUG_GLOBAL) << "[RendezvousProtocol] Loading ..." << endl;

	/* This is meant to be a singleton, so we will check if we have
	 * been loaded before. */
	if (protocolInstance)
	{
		kdDebug (RENDEZVOUS_DEBUG_GLOBAL) << "[RendezvousProtocol] Warning: Protocol already " << "loaded, not initializing again." << endl;
		return;
	}

	protocolInstance = this;

	// read the Rendezvous ID from Kopete's configuration
	KGlobal::config ()->setGroup ("Rendezvous");

	addAddressBookField ("messaging/p2pj", KopetePlugin::MakeIndexField);
}

RendezvousProtocol::~RendezvousProtocol ()
{
	//disconnectAll();

	/* make sure that the next attempt to load Rendezvous
	 * re-initializes the protocol class. */
	protocolInstance = 0L;
}



AddContactPage *RendezvousProtocol::createAddContactWidget (QWidget * parent, KopeteAccount * i)
{
	kdDebug (RENDEZVOUS_DEBUG_GLOBAL) << "[Rendezvous Protocol] Create Add Contact  Widget\n" << endl;;
	return new RendezvousAddContactPage (i, parent);
}

EditAccountWidget *RendezvousProtocol::createEditAccountWidget (KopeteAccount * account, QWidget * parent)
{
	kdDebug (RENDEZVOUS_DEBUG_GLOBAL) << "[Rendezvous Protocol] Edit Account Widget\n" << endl;
	return new RendezvousEditAccountWidget (this, static_cast < RendezvousAccount * >(account), parent);
}

KopeteAccount *RendezvousProtocol::createNewAccount (const QString & accountId)
{
	kdDebug (RENDEZVOUS_DEBUG_GLOBAL) << "[Rendezvous Protocol] Create New Account. ID: " << accountId << "\n" << endl;
	return new RendezvousAccount (this, accountId);
}


RendezvousProtocol *RendezvousProtocol::protocol ()
{
	// return current instance
	return protocolInstance;
}

void RendezvousProtocol::deserializeContact (KopeteMetaContact * metaContact,
										 const QMap < QString, QString > &serializedData, const QMap < QString, QString > & /* addressBookData */ )
{
//  kdDebug (RENDEZVOUS_DEBUG_GLOBAL) << k_funcinfo << "Deserializing data for metacontact " << metaContact->displayName () << "\n" << endl;

	QString contactId = serializedData["contactId"];
	QString displayName = serializedData["displayName"];
	QString accountId = serializedData["accountId"];

	QDict < KopeteAccount > accounts = KopeteAccountManager::manager ()->accounts (this);
	KopeteAccount *account = accounts[accountId];

	if (!account)
	{
		kdDebug(RENDEZVOUS_DEBUG_GLOBAL) << k_funcinfo << "WARNING: Account for contact does not exist, skipping." << endl;
		return;
	}

	if (account)
		account->addContact (contactId, displayName, metaContact);

}

#include "rendezvousprotocol.moc"
